"""A module defining functions to produce antenna positions algorithmically.

Each function here defined may take arbitrary parameters, but must return
a single array of shape (Nant, 3) with units of meters, corresponding to (x,y,z) positions
of antennae centred at zero.
"""

from __future__ import annotations

import numpy as np
from astropy import units as un

from . import units as tp
from . import yaml


@yaml.yaml_func()
def hera(
    hex_num: int,
    separation: tp.Length = 14 * un.m,
    split_core: bool = False,
    outriggers: bool = False,
    row_separation: tp.Length | None = None,
) -> tp.Meters:
    """
    Produce a simple regular hexagonal array.

    Note that the returned order of the core array antennas matches the antenna
    numbering used by the HERA team.

    Parameters
    ----------
    hex_num
        Number of antennas per side of the hexagon
    separation
        The distance between antennas along a side.
        May have units of distance or time, the latter interpreted as a distance travelled
        by light.
    split_core
        Whether to split the hex into three offset segments.
    outriggers
        Whether to add outrigger antennas.

    Returns
    -------
    antpos
        A 2D array of antenna positions, shape ``(Nants, 3)``.
    """
    if hex_num <= 1:
        raise ValueError("hex_num must be greater than 1")

    sep = separation.to_value("m")

    row_sep = sep * np.sqrt(3) / 2 if row_separation is None else row_separation.to_value("m")

    # construct the main hexagon
    positions = []
    for row in range(-hex_num + split_core + 1, hex_num):
        # adding split_core deletes a row if it's true
        for col in range(2 * hex_num - abs(row) - 1):
            x_pos = sep * ((2 - (2 * hex_num - abs(row))) / 2 + col)
            y_pos = row * row_sep
            positions.append([x_pos, y_pos, 0])

    # basis vectors (normalized to sep)
    up_right = sep * np.asarray([0.5, row_sep / sep, 0])
    up_left = sep * np.asarray([-0.5, row_sep / sep, 0])

    # split the core if desired
    if split_core:
        new_pos = []
        for pos in positions:
            # find out which sector the antenna is in
            theta = np.arctan2(pos[1], pos[0])
            if pos[0] == 0 and pos[1] == 0:
                new_pos.append(pos)
            elif -np.pi / 3 < theta < np.pi / 3:
                new_pos.append(np.asarray(pos) + (up_right + up_left) / 3)
            elif np.pi / 3 <= theta < np.pi:
                new_pos.append(np.asarray(pos) + up_left - (up_right + up_left) / 3)
            else:
                new_pos.append(pos)

        # update the positions
        positions = new_pos

    # add outriggers if desired
    if outriggers:
        # The specific displacements of the outrigger sectors are
        # designed specifically for redundant calibratability and
        # "complete" uv-coverage, but also to avoid specific
        # obstacles on the HERA site (e.g. a road to a MeerKAT antenna)
        exterior_hex_num = outriggers + 2
        for row in range(exterior_hex_num - 1, -exterior_hex_num, -1):
            for col in range(2 * exterior_hex_num - abs(row) - 1):
                x_pos = ((2 - (2 * exterior_hex_num - abs(row))) / 2 + col) * sep * (hex_num - 1)
                y_pos = row * (hex_num - 1) * row_sep
                theta = np.arctan2(y_pos, x_pos)
                if np.sqrt(x_pos**2 + y_pos**2) > sep * (hex_num + 1):
                    if 0 < theta <= 2 * np.pi / 3 + 0.01:
                        positions.append(
                            np.asarray([x_pos, y_pos, 0]) - 4 * (up_right + up_left) / 3
                        )
                    elif 0 >= theta > -2 * np.pi / 3:
                        positions.append(
                            np.asarray([x_pos, y_pos, 0]) - 2 * (up_right + up_left) / 3
                        )
                    else:
                        positions.append(
                            np.asarray([x_pos, y_pos, 0]) - 3 * (up_right + up_left) / 3
                        )

    return np.array(positions) * un.m



"""Test the profiles module."""

import astropy.units as un
import numpy as np
import pytest

from py21cmsense.observatory import Observatory, get_builtin_profiles


def test_get_available():
    avail = get_builtin_profiles()
    assert "HERA-H1C-IDR3" in avail


def test_load_hera_h1c():
    hera = Observatory.from_profile("HERA-H1C-IDR3")
    assert hera.Trcv == 100 * un.K
    assert hera.antpos.shape == (71, 3)


def test_load_hera_with_custom():
    hera = Observatory.from_profile("HERA-H1C-IDR3", Trcv=200 * un.K)
    assert hera.Trcv == 200 * un.K


all_profiles = [Observatory.from_profile(p) for p in get_builtin_profiles()]


def test_load_others():
    assert len({np.sum(o.antpos) for o in all_profiles}) == len(all_profiles)


def test_bad_profile():
    with pytest.raises(FileNotFoundError, match="profile invalid not available"):
        Observatory.from_profile("invalid")



"""Configuration for docs."""

import os
from datetime import datetime, timezone

from py21cmsense import __version__

extensions = [
    "sphinx.ext.autodoc",
    "sphinx.ext.autosummary",
    "sphinx.ext.coverage",
    "sphinx.ext.doctest",
    "sphinx.ext.extlinks",
    "sphinx.ext.ifconfig",
    "sphinx.ext.napoleon",
    "sphinx.ext.todo",
    "sphinx.ext.viewcode",
    "sphinx.ext.mathjax",
    "sphinx.ext.autosectionlabel",
    "numpydoc",
    "nbsphinx",
    "IPython.sphinxext.ipython_console_highlighting",
    "sphinx_design",
]
if os.getenv("SPELLCHECK"):
    extensions += ("sphinxcontrib.spelling",)
    spelling_show_suggestions = True
    spelling_lang = "en_US"

autosectionlabel_prefix_document = True

autosummary_generate = True
numpydoc_show_class_members = False

source_suffix = ".rst"
master_doc = "index"
project = "21cmSense"
year = str(datetime.now(tz=timezone.utc).year)
author = "Jonathan Pober and Steven Murray"
copyright = f"{year}, {author}"
version = release = __version__
templates_path = ["templates"]

pygments_style = "trac"
extlinks = {
    "issue": ("https://github.com/rasg-affiliates/21cmSense/issues/%s", "#"),
    "pr": ("https://github.com/rasg-affiliates/21cmSense/pull/%s", "PR #"),
}

html_theme = "furo"

html_use_smartypants = True
html_last_updated_fmt = "%b %d, %Y"
html_split_index = False
html_sidebars = {
    "**": [
        "sidebar/scroll-start.html",
        "sidebar/brand.html",
        "sidebar/search.html",
        "sidebar/navigation.html",
        "sidebar/ethical-ads.html",
        "sidebar/scroll-end.html",
    ]
}
html_short_title = f"{project}-{version}"

napoleon_use_ivar = True
napoleon_use_rtype = False
napoleon_use_param = False

mathjax_path = "https://cdn.jsdelivr.net/npm/mathjax@3/es5/tex-mml-chtml.js"

exclude_patterns = [
    "_build",
    "Thumbs.db",
    ".DS_Store",
    "templates",
    "**.ipynb_checkpoints",
]



"""Test the beam module."""

import numpy as np
import pytest
from astropy import units

from py21cmsense import GaussianBeam, beam


def test_abc():
    with pytest.raises(TypeError):
        beam.PrimaryBeam(150.0)


def test_gaussian_beam():
    bm = GaussianBeam(150.0 * units.MHz, dish_size=14 * units.m)

    assert bm.frequency == 150.0 * units.MHz
    assert bm.dish_size == 14 * units.m

    bm = GaussianBeam(0.15 * units.GHz, dish_size=1400 * units.cm)

    assert bm.frequency == 150.0 * units.MHz
    assert bm.dish_size == 14 * units.m

    assert not hasattr(bm.dish_size_in_lambda, "unit")
    assert bm.area != bm.at(160 * units.MHz).area

    with pytest.raises(NotImplementedError):
        GaussianBeam.from_uvbeam()

    assert bm.uv_resolution == bm.dish_size_in_lambda
    assert bm.sq_area < bm.area
    assert bm.fwhm > bm.width
    assert bm.first_null < np.pi * units.rad / 2
    assert bm.wavelength.unit == units.m

    assert bm.new() == bm



"""Test the theory module."""

import numpy as np
import pytest

from py21cmsense.theory import EOS2021, EOS2016Bright, EOS2016Faint, FarViewModel, Legacy21cmFAST


def test_eos_extrapolation():
    eos = EOS2021()

    with pytest.warns(UserWarning, match="Extrapolating above the simulated theoretical k"):
        eos.delta_squared(15, np.array([0.1, 1e6]))

    with pytest.warns(UserWarning, match="Extrapolating below the simulated theoretical k"):
        eos.delta_squared(15, np.array([0.0001, 0.1]))

    with pytest.warns(UserWarning, match="Extrapolating beyond simulated redshift range"):
        eos.delta_squared(50, np.array([0.1]))


def test_legacy():
    theory = Legacy21cmFAST()
    assert theory.delta_squared(9.1, 1.0) == theory.delta_squared(9.9, 1.0)

    with pytest.warns(UserWarning, match="Theory power corresponds to z=9.5, not z"):
        theory.delta_squared(1.0, 1.0)

    with pytest.warns(UserWarning, match="Extrapolating above the simulated theoretical k"):
        theory.delta_squared(9.5, np.array([0.1, 1e6]))

    with pytest.warns(UserWarning, match="Extrapolating below the simulated theoretical k"):
        theory.delta_squared(9.5, np.array([0.0001, 0.1]))


def test_eos_2016():
    faint = EOS2016Faint()
    bright = EOS2016Bright()

    assert faint.delta_squared(9.1, 1.0) != bright.delta_squared(9.1, 1.0)


def test_FarView():
    theory = FarViewModel()
    assert theory.delta_squared(29.6, 1.0) == theory.delta_squared(30.4, 1.0)

    with pytest.warns(UserWarning, match="Theory power corresponds to z=30, not z"):
        theory.delta_squared(1.0, 1.0)

    with pytest.warns(UserWarning, match="Extrapolating above the simulated theoretical k"):
        theory.delta_squared(30, np.array([0.1, 1e6]))

    with pytest.warns(UserWarning, match="Extrapolating below the simulated theoretical k"):
        theory.delta_squared(30, np.array([0.0001, 0.1]))



"""Test the yaml module."""

import pickle

import numpy as np
import pytest
from astropy import units as un
from astropy.io.misc import yaml

from py21cmsense.yaml import LoadError


def test_file_not_found():
    with pytest.raises(IOError, match="not found"):
        yaml.load("!txt non-existent-file.txt")


def test_bad_loader(tmpdirec):
    fl = tmpdirec / "test-npy"

    with open(fl, "w") as ff:
        ff.write("some-text")

    with pytest.raises(LoadError):
        yaml.load(f"!npy {fl}")


def test_pickle_loader(tmpdirec):
    pkl = tmpdirec / "test-pkl.pkl"

    obj = {"an": "object"}
    with open(pkl, "wb") as fl:
        pickle.dump(obj, fl)

    d = yaml.load(f"!pkl {pkl}")

    assert d == obj


def test_npz_loader(tmpdirec):
    npz = tmpdirec / "test-npz.npz"

    obj = {"an": np.linspace(0, 1, 10), "b": np.zeros(10)}

    np.savez(npz, **obj)

    d = yaml.load(f"!npz {npz}")

    for k, v in d.items():
        assert k in obj
        assert np.allclose(v, obj[k])


def test_txt_loader_with_unit(tmpdirec):
    txt = tmpdirec / "test-txt.txt"

    obj = np.linspace(0, 1, 10)

    np.savetxt(txt, obj)

    d = yaml.load(f"!txt {txt} | m")
    assert d.unit == un.m
    assert np.allclose(d, obj * un.m)



"""
Module providing the definition of an Observatory.

This replaces the original usage of an aipy.AntennaArray with something much more
simple, and suited to the needs of this particular package.
"""

from __future__ import annotations

import collections
import logging
from collections import defaultdict
from functools import cached_property
from pathlib import Path
from typing import Callable

import attr
import numpy as np
import tqdm
from astropy import constants as cnst
from astropy import units as un
from astropy.io.misc import yaml
from attr import validators as vld
from hickleable import hickleable

from . import _utils as ut
from . import beam, config
from . import units as tp

logger = logging.getLogger(__name__)

DATA = Path(__file__).parent / "data"


def get_builtin_profiles() -> list[str]:
    """Print available built-in profiles."""
    fls = (DATA / "profiles").glob("*.yaml")
    return [fl.stem for fl in fls]


@hickleable(evaluate_cached_properties=True)
@attr.s(kw_only=True, order=False)
class Observatory:
    """
    A class defining an interferometric Observatory and its properties.

    Parameters
    ----------
    antpos : array
        An array with shape (Nants, 3) specifying the positions of the antennas.
        These should be in the ENU (East-North-Up) frame, relative to a central location
        given by `latitude`. If not a Quantity, units are assumed to be meters.
    beam : :class:`~py21cmsense.beam.PrimaryBeam` instance
        A beam, assumed to be homogeneous across antennas.
    latitude : float or Quantity, optional
        Latitude of the array center. If a float, assumed to be in radians.
        Note that longitude is not required, as we assume an isotropic sky.
    Trcv
        Receiver temperature, either a temperature Quantity, or a callable that
        takes a single frequency Quantity and returns a temperature Quantity.
    min_antpos, max_antpos
        The minimum/maximum radial distance to include antennas (from the origin
        of the array). Assumed to be in units of meters if no units are supplied.
        Can be used to limit antennas in arrays like HERA and SKA that
        have a "core" and "outriggers". The minimum is inclusive, and maximum exclusive.
    beam_crossing_time_incl_latitude
        Whether the beam-crossing time is dependent on the latitude of the telescope.
        By default it is, so that the beam-crossing time is
        ``tday * FWHM / (2pi cos(lat))``. This affects both the thermal and sample
        variance calculations.
    world: string
        A string specifying whether the telescope is on the Earth or the moon.
    """

    _antpos: tp.Length = attr.ib(eq=attr.cmp_using(eq=np.array_equal))
    beam: beam.PrimaryBeam = attr.ib(validator=vld.instance_of(beam.PrimaryBeam))
    latitude: un.rad = attr.ib(
        0 * un.rad,
        validator=ut.between(-np.pi * un.rad / 2, np.pi * un.rad / 2),
    )
    Trcv: tp.Temperature | Callable = attr.ib(100 * un.K)
    max_antpos: tp.Length = attr.ib(
        default=np.inf * un.m, validator=(tp.vld_physical_type("length"), ut.positive)
    )
    min_antpos: tp.Length = attr.ib(
        default=0.0 * un.m, validator=(tp.vld_physical_type("length"), ut.nonnegative)
    )
    beam_crossing_time_incl_latitude: bool = attr.ib(default=True, converter=bool)
    world: str = attr.ib(default="earth", validator=vld.in_(["earth", "moon"]))

    @_antpos.validator
    def _antpos_validator(self, att, val):
        tp.vld_physical_type("length")(self, att, val)
        if val.ndim != 2:
            raise ValueError("antpos must be a 2D array.")

        if val.shape[-1] != 3:
            raise ValueError("antpos must have shape (Nants, 3).")

        if val.shape[0] <= 1:
            raise ValueError("antpos must have at least two antennas.")

    @cached_property
    def antpos(self) -> np.ndarray:
        """The positions of antennas in the array in units of metres."""
        # Mask out some antennas if a max_antpos is set in the YAML
        _n = len(self._antpos)
        sq_len = np.sum(np.square(self._antpos), axis=1)
        antpos = self._antpos[
            np.logical_and(
                sq_len >= self.min_antpos**2,
                sq_len < self.max_antpos**2,
            )
        ]

        if self.max_antpos < np.inf or self.min_antpos > 0:
            logger.info(
                f"Removed {_n - len(antpos)} antennas using given "
                f"max_antpos={self.max_antpos} m and min_antpos={self.min_antpos} m."
            )

        return antpos

    @Trcv.validator
    def _trcv_vld(self, att, val):
        if callable(val):
            try:
                y = val(1 * un.MHz)
            except Exception as e:
                raise ValueError(
                    "Trcv function must take a frequency Quantity and "
                    "return a temperature Quantity."
                ) from e

            if not (isinstance(y, un.Quantity) and y.unit.physical_type == "temperature"):
                raise ValueError("Trcv function must return a temperature Quantity.")
        else:
            tp.vld_physical_type("temperature")(self, att, val)

    @property
    def frequency(self) -> un.Quantity[un.MHz]:
        """Central frequency of the observation."""
        return self.beam.frequency.to("MHz")

    @cached_property
    def n_antennas(self) -> int:
        """Number of antennas in the array."""
        return len(self.antpos)

    def clone(self, **kwargs) -> Observatory:
        """Return a clone of this instance, but change kwargs."""
        return attr.evolve(self, **kwargs)

    @classmethod
    def from_uvdata(cls, uvdata, beam: beam.PrimaryBeam, **kwargs) -> Observatory:
        """Instantiate an Observatory from a :class:`pyuvdata.UVData` object."""
        return cls(
            antpos=uvdata.telescope.antenna_positions,
            beam=beam,
            latitude=uvdata.telescope.location.lat,
            **kwargs,
        )

    @classmethod
    def from_yaml(cls, yaml_file: str | dict, frequency: tp.Frequency | None = None) -> Observatory:
        """Instantiate an Observatory from a compatible YAML config file."""
        if isinstance(yaml_file, (str, Path)):
            with open(yaml_file) as fl:
                lines = fl.read()
                lines = lines.replace("{{ DATA_PATH }}", str(DATA.absolute()))
                data = yaml.load(lines)
        elif isinstance(yaml_file, collections.abc.Mapping):
            data = yaml_file
        else:
            raise ValueError("yaml_file must be a string filepath or a raw dict from such a file.")

        # Mask out some antennas if a max_antpos is set in the YAML
        max_antpos = data.pop("max_antpos", np.inf * un.m)
        antpos = data.pop("antpos")
        _n = len(antpos)

        antpos = antpos[np.sum(np.square(antpos), axis=1) < max_antpos**2]

        if max_antpos < np.inf * un.m:
            logger.info(
                f"Removed {_n - len(antpos)} antennas using given max_antpos={max_antpos} m."
            )

        # If we get only East and North coords, add zeros for the UP direction.
        if antpos.shape[1] == 2:
            antpos = np.hstack((antpos, np.zeros((len(antpos), 1))))

        _beam = data.pop("beam")
        if frequency is not None:
            _beam["frequency"] = frequency

        kind = _beam.pop("class")
        _beam = getattr(beam, kind)(**_beam)

        return cls(antpos=antpos, beam=_beam, **data)

    @classmethod
    def from_profile(cls, profile: str, frequency: tp.Frequency | None = None, **kwargs):
        """Instantiate the Observatory from a builtin profile.

        Parameters
        ----------
        profile
            A string label identifying the observatory. Available built-in observatories
            can be obtained with :func:`get_builtin_profiles`. For more up-to-date SKA profiles,
            check the :func:`from_ska` method.
        frequency
            The frequency at which to specify the observatory.

        Other Parameters
        ----------------
        All other parameters passed will be passed into the initializer for the class,
        overwriting the profile.
        """
        fl = DATA / "profiles" / f"{profile}.yaml"
        if not fl.exists():
            raise FileNotFoundError(
                f"profile {profile} not available. Available profiles: {get_builtin_profiles()}"
            )

        obj = cls.from_yaml(fl, frequency=frequency)
        return obj.clone(**kwargs)

    @classmethod
    def from_ska(
        cls,
        subarray_type: str,
        array_type: str = "low",
        Trcv: tp.Temperature | Callable = 100 * un.K,  # noqa N803
        frequency: tp.Frequency | None = 150.0 * un.MHz,
        **kwargs,
    ) -> Observatory:
        """Instantiate an SKA Observatory.

        Parameters
        ----------
        subarray_type
            The type of subarray to use. Options are "AA4", "AA*", "AA1", "AA2", "AA0.5",
             and "custom"
        array_type, optional
            The type of array to use. Options are "low" and "mid".
            Default is "low".
        Trcv, optional
            Receiver temperature, either a temperature Quantity, or a callable that
            takes a single frequency Quantity and returns a temperature Quantity.
            Default is 100 K.
        frequency, optional
            The frequency at which to specify the observatory. Default is 150 MHz.

        Other Parameters
        ----------------
        All other parameters passed will be passed into the LowSubArray or MidSubArray class.
        See the documentation of the ska-ost-array-config package for more information.
        """
        try:
            from ska_ost_array_config.array_config import LowSubArray, MidSubArray
        except ImportError as exception:  # pragma: no cover
            raise ImportError(
                "ska-ost-array-config package is required, "
                + "see https://gitlab.com/ska-telescope/ost/ska-ost-array-config"
            ) from exception

        if array_type == "low":
            subarray = LowSubArray(subarray_type, **kwargs)
        elif array_type == "mid":
            subarray = MidSubArray(subarray_type, **kwargs)
        else:
            raise ValueError("array_type must be 'low' or 'mid'.")
        antpos = subarray.array_config.xyz.data * un.m
        _beam = beam.GaussianBeam(
            frequency=frequency, dish_size=np.array(subarray.array_config.diameter)[0] * un.m
        )
        lat = subarray.array_config.location.lat.rad * un.rad
        return cls(antpos=antpos, beam=_beam, latitude=lat, Trcv=Trcv)

    @cached_property
    def baselines_metres(self) -> tp.Meters:
        """Raw baseline distances in metres for every pair of antennas.

        Shape is ``(Nant, Nant, 3)``.
        """
        # this does an "outer" subtraction, leaving the inner 2- or 3- length positions
        # as atomic quantities.
        return (self.antpos[np.newaxis, :, :] - self.antpos[:, np.newaxis, :]).to(un.m)

    def projected_baselines(
        self,
        baselines: tp.Length | None = None,
        time_offset: tp.Time = 0 * un.hour,
        phase_center_dec: tp.Angle | None = None,
    ) -> np.ndarray:
        """Compute the *projected* baseline lengths (in wavelengths).

        Phased to a point that has rotated off zenith by some time_offset.

        Parameters
        ----------
        baselines
            The baseline co-ordinates to project, assumed to be in metres.
            If not provided, uses all baselines of the observatory.
            Shape of the array can be (N,N,3) or (N, 3).
            The co-ordinates are expected to be in ENU.
        time_offset
            The amount of time elapsed since the phase center was at zenith.
            Assumed to be in days unless otherwise defined. May be negative.
        phase_center_dec
            The declination of the phase center of the observation. By default, the
            same as the latitude of the array.

        Returns
        -------
        An array the same shape as :attr:`baselines_metres`, but phased to the
        new phase centre.
        """
        if baselines is None:
            baselines = self.baselines_metres

        orig_shape = baselines.shape

        bl_wavelengths = baselines.reshape((-1, 3)) * self.metres_to_wavelengths

        out = ut.phase_past_zenith(
            time_past_zenith=time_offset,
            bls_enu=bl_wavelengths,
            latitude=self.latitude,
            world=self.world,
            phase_center_dec=phase_center_dec,
        )

        out = out.reshape(*orig_shape[:-1], np.size(time_offset), orig_shape[-1])
        if np.size(time_offset) == 1:
            out = out.squeeze(-2)

        return out

    @cached_property
    def metres_to_wavelengths(self) -> un.Quantity[1 / un.m]:
        """Conversion factor for metres to wavelengths at fiducial frequency."""
        return (self.frequency / cnst.c).to("1/m")

    @cached_property
    def baseline_lengths(self) -> np.ndarray:
        """Lengths of baselines in units of wavelengths, shape (Nant, Nant)."""
        return np.sqrt(np.sum(self.projected_baselines() ** 2, axis=-1))

    @cached_property
    def shortest_baseline(self) -> float:
        """Shortest baseline in units of wavelengths."""
        return np.min(self.baseline_lengths[self.baseline_lengths > 0])

    @cached_property
    def longest_baseline(self) -> float:
        """Longest baseline in units of wavelengths."""
        return np.max(self.baseline_lengths)

    @cached_property
    def observation_duration(self) -> un.Quantity[un.day]:
        """The time it takes for the sky to drift through the FWHM."""
        latfac = np.cos(self.latitude) if self.beam_crossing_time_incl_latitude else 1
        if self.world == "earth":
            return un.day * self.beam.fwhm / (2 * np.pi * un.rad * latfac)
        else:
            return 27.3 * un.day * self.beam.fwhm / (2 * np.pi * un.rad * latfac)

    def get_redundant_baselines(
        self,
        baseline_filters: Callable | tuple[Callable] = (),
        ndecimals: int = 1,
    ) -> dict[tuple[float, float, float], list[tuple[int, int]]]:
        """
        Determine all baseline groups.

        Parameters
        ----------
        baseline_filters
            Callable function (or functions) of a single 3-coordinate baseline vector
            that returns a bool indicating whether to include the baseline.
        ndecimals : int, optional
            The number of decimals to which the UV points must be the same to be
            considered redundant.

        Returns
        -------
        dict: a dictionary in which keys are 3-tuples of ``(u,v, |u|)`` co-ordinates and
            values are lists of 2-tuples, where each 2-tuple consists of the indices
            of a pair of antennas with those co-ordinates.
        """
        uvbins = defaultdict(list)
        baseline_filters = tp._tuplify(baseline_filters, 1)

        def filt(blm):
            return all(filt(blm) for filt in baseline_filters)

        # Everything here is in wavelengths
        uvw = self.projected_baselines()[:, :, :2].value
        uvw = np.round(uvw, decimals=ndecimals)
        bl_lens = np.round(self.baseline_lengths.value, decimals=ndecimals)

        # group redundant baselines
        for i in tqdm.tqdm(
            range(self.n_antennas - 1),
            desc="finding redundancies",
            unit="ants",
            disable=not config.PROGRESS,
        ):
            for j in range(i + 1, self.n_antennas):
                blm = self.baselines_metres[i, j]

                # Check if we want to include this baseline.
                if not filt(blm):
                    continue

                bl_len = bl_lens[i, j]  # in wavelengths

                u, v = uvw[i, j]

                # add the uv point and its inverse to the redundant baseline dict.
                uvbins[(u, v, bl_len)].append((i, j))
                uvbins[(-u, -v, bl_len)].append((j, i))

        return uvbins

    def time_offsets_from_obs_int_time(
        self, integration_time: tp.Time, observation_duration: tp.Time | None = None
    ):
        """Compute a list of time offsets within an LST-bin.

        The LSTs 'within a bin' are added coherently for a given baseline group.
        Time offsets are with respect to an arbitrary time, and describe the rotation of
        a hypothetical point through zenith.

        Parameters
        ----------
        integration_time
            Time for single snapshot.
        observation_duration
            Duration of the LST bin (for single night).

        Returns
        -------
        array :
            Time offsets (in julian days).
        """
        if observation_duration is None:
            observation_duration = self.observation_duration

        assert integration_time <= observation_duration

        return (
            np.arange(
                -observation_duration.to("day").value / 2,
                observation_duration.to("day").value / 2,
                integration_time.to("day").value,
            )
            << un.day
        )

    def baseline_coords_from_groups(self, baseline_groups) -> un.Quantity[un.m]:
        """Convert a dictionary of baseline groups to an array of ENU co-ordinates."""
        out = np.zeros((len(baseline_groups), 3)) * un.m
        for i, antpairs in enumerate(baseline_groups.values()):
            out[i] = self.baselines_metres[antpairs[0][0], antpairs[0][1]]
        return out

    @staticmethod
    def baseline_weights_from_groups(baseline_groups) -> np.ndarray:
        """Get number of baselines in each group.

        Parameters
        ----------
        baseline_groups
            A dictionary in the format output by :func:`get_redundant_baselines`.

        Returns
        -------
        weights
            An array containing the number of baselines in each group.
        """
        return np.array([len(antpairs) for antpairs in baseline_groups.values()])

    def grid_baselines(
        self,
        coherent: bool,
        baselines: tp.Length | None = None,
        weights: np.ndarray | None = None,
        integration_time: tp.Time = 60.0 * un.s,
        baseline_filters: Callable | tuple[Callable] = (),
        observation_duration: tp.Time | None = None,
        ndecimals: int = 1,
        phase_center_dec: tp.Angle | None = None,
    ) -> np.ndarray:
        """
        Grid baselines onto a pre-determined uvgrid, accounting for earth rotation.

        Parameters
        ----------
        baselines : array_like, optional
            The baseline co-ordinates to project, assumed to be in metres.
            If not provided, calculates effective baselines by finding redundancies on
            all baselines in the observatory. Shape of the array can be (N,N,3) or (N, 3).
            The co-ordinates are expected to be in ENU. If `baselines` is provided,
            `weights` must also be provided.
        weights: array_like, optional
            An array of the same length as `baselines`, giving the number of independent
            baselines at each co-ordinate. If not provided, calculates effective
            baselines by finding redundancies on all baselines in the observatory.
            If `baselines` is provided, `weights` must also be provided.
        integration_time : float or Quantity, optional
            The amount of time integrated into a snapshot visibility, assumed
            to be in seconds.
        baseline_filters
            A function that takes a single value: a length-3 array of baseline co-ordinates,
            and returns a bool indicating whether to include the baseline. Built-in filters
            are provided in the :mod:`~baseline_filters` module.
        observation_duration : float or Quantity, optional
            Amount of time in a single (coherent) LST bin, assumed to be in minutes.
        ndecimals : int, optional
            Number of decimals to which baselines must match to be considered redundant.

        Returns
        -------
        array :
            Shape [n_baseline_groups, Nuv, Nuv]. The coherent sum of baselines within
            grid cells given by :attr:`ugrid`. One can treat different baseline groups
            independently, or sum over them.

        See Also
        --------
        grid_baselines_coherent :
            Coherent sum over baseline groups of the output of this method.
        grid_baseline_incoherent :
            Incoherent sum over baseline groups of the output of this method.
        """
        if baselines is not None:
            assert un.get_physical_type(baselines) == "length"
            assert baselines.ndim in (2, 3)

        assert un.get_physical_type(integration_time) == "time"

        if observation_duration is not None:
            assert un.get_physical_type(observation_duration) == "time"

        if baselines is None:
            baseline_groups = self.get_redundant_baselines(
                baseline_filters=baseline_filters, ndecimals=ndecimals
            )
            baselines = self.baseline_coords_from_groups(baseline_groups)
            weights = self.baseline_weights_from_groups(baseline_groups)

        bl_max = np.sqrt(np.max(np.sum(baselines**2, axis=1)))

        if weights is None:
            raise ValueError("If baselines are provided, weights must also be provided.")

        time_offsets = self.time_offsets_from_obs_int_time(integration_time, observation_duration)

        uvws = self.projected_baselines(
            baselines, time_offsets, phase_center_dec=phase_center_dec
        ).reshape(baselines.shape[0], time_offsets.size, 3)

        # grid each baseline type into uv plane
        dim = len(self.ugrid(bl_max))
        edges = self.ugrid_edges(bl_max)

        uvsum = np.zeros((dim, dim))
        for uvw, nbls in tqdm.tqdm(
            zip(uvws, weights),
            desc="gridding baselines",
            unit="baselines",
            disable=not config.PROGRESS,
            total=len(weights),
        ):
            hist = np.histogram2d(uvw[:, 0], uvw[:, 1], bins=edges)[0] * nbls

            uvsum += hist if coherent else hist**2
        if not coherent:
            uvsum = np.sqrt(uvsum)

        return uvsum

    def longest_used_baseline(self, bl_max: tp.Length = np.inf * un.m) -> float:
        """Determine the maximum baseline length kept in the array, in wavelengths."""
        if np.isinf(bl_max):
            return self.longest_baseline

        # Note we don't do the conversion in-place!
        bl_max = bl_max * self.metres_to_wavelengths
        return np.max(self.baseline_lengths[self.baseline_lengths <= bl_max])

    def ugrid_edges(self, bl_max: tp.Length = np.inf * un.m) -> np.ndarray:
        """Get a uv grid out to the maximum used baseline smaller than given bl_max.

        The resulting array represents the *edges* of the grid (so the number of cells
        is one fewer than this).

        Parameters
        ----------
        bl_max : float or Quantity
            Include all baselines smaller than this number. Units of m.

        Returns
        -------
        array :
            1D array of regularly spaced un.
        """
        bl_max = self.longest_used_baseline(bl_max)

        # We're doing edges of bins here, and the first edge is at uv_res/2
        n_positive = int(np.ceil((bl_max - self.beam.uv_resolution / 2) / self.beam.uv_resolution))

        # Grid from uv_res/2 to just past (or equal to) bl_max, in steps of resolution.
        positive = np.linspace(
            self.beam.uv_resolution / 2,
            self.beam.uv_resolution / 2 + n_positive * self.beam.uv_resolution,
            n_positive + 1,
        )
        return np.concatenate((-positive[::-1], positive))

    def ugrid(self, bl_max: tp.Length = np.inf * un.m) -> np.ndarray:
        """Centres of the UV grid plane."""
        # Shift the edges by half a cell, and omit the last one
        edges = self.ugrid_edges(bl_max)
        return (edges[1:] + edges[:-1]) / 2



"""Test the antenna positions."""

import numpy as np
import pytest
from astropy import units as un

from py21cmsense.antpos import hera


@pytest.mark.parametrize("n", [3, 5, 8])
def test_hera_split_core(n):
    # https://en.wikipedia.org/wiki/Centered_hexagonal_number
    # 3*n^2 - 3n + 1
    antpos1 = hera(hex_num=n)
    assert len(antpos1) == 3 * n**2 - 3 * n + 1
    antpos2 = hera(hex_num=n, split_core=True)
    assert len(antpos2) == len(antpos1) - n


@pytest.mark.parametrize("n", [3, 5, 8])
def test_hera_outriggers(n):
    # https://en.wikipedia.org/wiki/Centered_hexagonal_number
    # 3*n^2 - 3n + 1
    antpos1 = hera(hex_num=n)
    assert len(antpos1) == 3 * n**2 - 3 * n + 1
    antpos2 = hera(hex_num=n, outriggers=1)
    assert len(antpos2) >= len(antpos1)


def test_hera_set_row_sep():
    antpos1 = hera(4)
    antpos2 = hera(4, row_separation=14 * np.sin(np.pi / 3) * un.m)

    assert np.allclose(antpos1, antpos2)

    antpos3 = hera(4, row_separation=12.12 * un.m)
    assert not np.allclose(antpos1, antpos3)


def test_bad_hex_num():
    with pytest.raises(ValueError, match="hex_num must be greater than 1"):
        hera(1)



"""A module defining an interface for theoretical predictions, eg. power spectra.

This is simply so that one can calculate sample variance values for a particular
theoretical model, with a uniform interface.

To register a class as a theory power spectrum, simply make it a subclass of
``TheoryModel``. You must define two things on the class: (1) a class attribute
"use_littleh" which says whether the function that evaluates the power spectrum expects
the wavenumber in h/Mpc or 1/Mpc, and (2) a ``__call__`` method, which takes in a float
redshift and an array of wavenumbers, and returns Delta^2 as an astropy Quantity with
units mK^2.
"""

import abc
import warnings
from pathlib import Path

import numpy as np
from astropy import units as un
from scipy.interpolate import InterpolatedUnivariateSpline, RectBivariateSpline

_ALL_THEORY_POWER_SPECTRA = {}


class TheoryModel(abc.ABC):
    """Abstract base class for theory models.

    Subclasses must implement the :meth:`delta_squared` method.
    """

    #: Whether the theory model uses little-h units for wavenumbers.
    use_littleh: bool = False

    def __init_subclass__(cls) -> None:
        """Add the subclass to the plugin dict."""
        _ALL_THEORY_POWER_SPECTRA[cls.__name__] = cls
        return super().__init_subclass__()

    @abc.abstractmethod
    def delta_squared(self, z: float, k: np.ndarray) -> un.Quantity[un.mK**2]:
        """Compute Delta^2(k, z) for the theory model.

        Parameters
        ----------
        z
            The redshift (should be a float).
        k
            The wavenumbers, either in units of 1/Mpc if use_littleh=False, or
            h/Mpc if use_littleh=True.

        Returns
        -------
        delta_squared
            An array of delta_squared values in units of mK^2.
        """
        # pragma: no cover


class TheorySpline(TheoryModel):
    """Base class for theory models that are defined by a spline over (z,k)."""

    def delta_squared(self, z: float, k: np.ndarray) -> un.Quantity[un.mK**2]:
        """Compute Delta^2(k, z) for the theory model.

        Parameters
        ----------
        z
            The redshift (should be a float).
        k
            The wavenumbers, either in units of 1/Mpc if use_littleh=False, or
            h/Mpc if use_littleh=True.

        Returns
        -------
        delta_squared
            An array of delta_squared values in units of mK^2.
        """
        if np.any(k > self.k.max()):
            warnings.warn(
                f"Extrapolating above the simulated theoretical k: {k.max()} > {self.k.max()}",
                stacklevel=2,
            )
        if np.any(k < self.k.min()):
            warnings.warn(
                f"Extrapolating below the simulated theoretical k: {k.min()} < {self.k.min()}",
                stacklevel=2,
            )
        if not self.z.min() <= z <= self.z.max():
            warnings.warn(
                f"Extrapolating beyond simulated redshift range: {z} not in "
                f"range ({self.z.min(), self.z.max()})",
                stacklevel=2,
            )

        return self.spline(z, k, grid=False) << un.mK**2


class EOS2021(TheorySpline):
    """Theory model from EOS2021 (https://arxiv.org/abs/2110.13919)."""

    use_littleh = False

    def __init__(self):
        pth = Path(__file__).parent / "data/eos2021"
        z = np.fromfile(pth / "1pt5Gpc_EOS_coeval_pow_zlist.bin")
        # TODO: we divide by 2.5 here as the k values on the EOS2021 GDrive are wrong --
        # they are for the 600 Mpc box instead of the 1.5 Gpc box. Later when that's
        # fixed we should just fix the data here.
        self.k = np.fromfile(pth / "1pt5Gpc_EOS_coeval_pow_kbins.bin") / 2.5
        coeval_ps = np.fromfile(pth / "1pt5Gpc_EOS_coeval_pow_P21.bin").reshape(
            (z.size, self.k.size)
        )

        # Sort in order of ascending redshift
        order = np.argsort(z)
        self.z = z[order]
        self.coeval_ps = coeval_ps[order]

        self.spline = RectBivariateSpline(self.z, self.k, self.coeval_ps, ky=1)


def _get_eos2016_data(name: str):
    pth = Path(__file__).parent / f"data/{name}"
    all_files = sorted(pth.glob("ps_no_halos*"))
    z = np.array([float(f.name.split("_")[3][1:]) for f in all_files])
    k = np.genfromtxt(all_files[0])[:, 0]

    coeval_ps = np.array([np.genfromtxt(fl)[:, 1] for fl in all_files])

    spline = RectBivariateSpline(z, k, coeval_ps, ky=1)
    return z, k, coeval_ps, spline


class EOS2016Faint(TheorySpline):
    """Theory model from EOS2016 Faint Galaxies."""

    use_littleh = False

    def __init__(self):
        self.z, self.k, self.coeval_ps, self.spline = _get_eos2016_data("eos16-faint")


class EOS2016Bright(TheorySpline):
    """Theory model from EOS2016 Bright Galaxies."""

    use_littleh = False

    def __init__(self):
        self.z, self.k, self.coeval_ps, self.spline = _get_eos2016_data("eos16-bright")


class Legacy21cmFAST(TheoryModel):
    """Simple 21cmFAST-based theory model explicitly for z=9.5, from 21cmSense v1."""

    use_littleh: bool = False

    def __init__(self) -> None:
        pth = (
            Path(__file__).parent
            / "data/ps_no_halos_nf0.521457_z9.50_useTs0_zetaX-1.0e+00_200_400Mpc_v2"
        )
        self.k, self.delta_squared_raw = np.genfromtxt(pth).T[:2]

        self.spline = InterpolatedUnivariateSpline(self.k, self.delta_squared_raw, k=1)

    def delta_squared(self, z: float, k: np.ndarray) -> un.Quantity[un.mK**2]:
        """Compute Delta^2(k, z) for the theory model.

        Parameters
        ----------
        z
            The redshift (should be a float).
        k
            The wavenumbers, either in units of 1/Mpc if use_littleh=False, or
            h/Mpc if use_littleh=True.

        Returns
        -------
        delta_squared
            An array of delta_squared values in units of mK^2.
        """
        if np.any(k > self.k.max()):
            warnings.warn(
                f"Extrapolating above the simulated theoretical k: {k.max()} > {self.k.max()}",
                stacklevel=2,
            )
        if np.any(k < self.k.min()):
            warnings.warn(
                f"Extrapolating below the simulated theoretical k: {k.min()} < {self.k.min()}",
                stacklevel=2,
            )
        if not 9 < z < 10:
            warnings.warn(
                f"Theory power corresponds to z=9.5, not z={z:.2f}",
                stacklevel=2,
            )

        return self.spline(k) << un.mK**2


class FarViewModel(TheoryModel):
    """21cmFAST-based theory model explicitly for z=30."""

    use_littleh: bool = False

    def __init__(self) -> None:
        k_pth = Path(__file__).parent / "data/farview/kmag_ultimate.npy"

        delta_pth = Path(__file__).parent / "data/farview/P_Tb_ultimate.npy"
        # Should at some point reorganize the data so these steps aren't necessary
        k_fixed = np.load(k_pth)
        power = np.load(delta_pth)
        k_fixed = k_fixed[~np.isnan(power)]
        power = power[~np.isnan(power)]
        delta = (k_fixed**3 * power) / (2 * np.pi**2)

        self.k = k_fixed
        self.delta_squared_raw = delta

        self.spline = InterpolatedUnivariateSpline(self.k, self.delta_squared_raw, k=1)

    def delta_squared(self, z: float, k: np.ndarray) -> un.Quantity[un.mK**2]:
        """Compute Delta^2(k, z) for the theory model.

        Parameters
        ----------
        z
            The redshift (should be a float).
        k
            The wavenumbers, either in units of 1/Mpc if use_littleh=False, or
            h/Mpc if use_littleh=True.

        Returns
        -------
        delta_squared
            An array of delta_squared values in units of mK^2.
        """
        if np.any(k > self.k.max()):
            warnings.warn(
                f"Extrapolating above the simulated theoretical k: {k.max()} > {self.k.max()}",
                stacklevel=2,
            )
        if np.any(k < self.k.min()):
            warnings.warn(
                f"Extrapolating below the simulated theoretical k: {k.min()} < {self.k.min()}",
                stacklevel=2,
            )
        if not 29.5 < z < 30.5:
            warnings.warn(
                f"Theory power corresponds to z=30, not z={z:.2f}",
                stacklevel=2,
            )

        return self.spline(k) << un.mK**2



"""Test the units module."""

import attr
import pytest
from astropy import units as u

from py21cmsense import units as tp


def test_vld_physical_type():
    @attr.s
    class K:
        a = attr.ib(validator=tp.vld_physical_type("frequency"))

    with pytest.raises(tp.UnitError, match="a must be an astropy Quantity!"):
        K(1)

    with pytest.raises(
        u.UnitConversionError,
        match="a must have physical type of 'frequency'. Got 'time'",
    ):
        K(1 * u.s)

    assert K(1 * u.MHz).a.unit.physical_type == "frequency"


def test_vld_unit():
    @attr.s
    class K:
        a = attr.ib(validator=tp.vld_unit(u.GHz))

    with pytest.raises(tp.UnitError, match="a must be an astropy Quantity!"):
        K(1)

    with pytest.raises(u.UnitConversionError, match="a not convertible to GHz. Got s"):
        K(1 * u.s)

    assert K(1 * u.MHz).a.unit.physical_type == "frequency"


def test_tuplify():
    x = tp._tuplify(1)
    assert x == (1,)

    x = tp._tuplify([1])
    assert x == (1,)

    x = tp._tuplify((1,))
    assert x == (1,)

    x = tp._tuplify((1, 2))
    assert x == (1, 2)

    x = tp._tuplify(None, n=3)
    assert x == (None, None, None)



"""Test the observatory module."""

import re
from pathlib import Path

import numpy as np
import pytest
import pyuvdata
from astropy import units
from astropy.coordinates import EarthLocation, SkyCoord
from astropy.time import Time

from py21cmsense import Observatory
from py21cmsense.baseline_filters import BaselineRange
from py21cmsense.beam import GaussianBeam
from py21cmsense.data import PATH


@pytest.fixture(scope="module")
def bm():
    return GaussianBeam(frequency=150.0 * units.MHz, dish_size=14 * units.m)


def test_antpos(bm):
    a = Observatory(antpos=np.zeros((10, 3)) * units.m, beam=bm)
    assert a.antpos.unit == units.m

    assert np.all(a.baselines_metres == 0)

    # If bad units given, should raise error.
    with pytest.raises(units.UnitConversionError):
        Observatory(antpos=np.zeros((10, 3)) * units.s, beam=bm)

    # Need more than one antenna
    with pytest.raises(ValueError, match="antpos must have at least two antennas"):
        Observatory(antpos=np.zeros((1, 3)) * units.m, beam=bm)

    with pytest.raises(ValueError, match="antpos must be a 2D array"):
        Observatory(antpos=np.zeros(10) * units.m, beam=bm)

    with pytest.raises(ValueError, match=re.escape("antpos must have shape (Nants, 3)")):
        Observatory(antpos=np.zeros((10, 2)) * units.m, beam=bm)


def test_observatory_class(bm):
    a = Observatory(antpos=np.zeros((3, 3)) * units.m, beam=bm)
    b = a.clone()
    assert a == b


def test_Trcv(bm):
    a = Observatory(antpos=np.zeros((3, 3)) * units.m, beam=bm, Trcv=10 * units.mK)
    assert a.Trcv.unit == units.mK


def test_Trcv_func(bm):
    a = Observatory(
        antpos=np.zeros((3, 3)) * units.m,
        beam=bm,
        Trcv=lambda f: (f / units.MHz) * 10 * units.mK,
    )
    assert a.Trcv(7 * units.Hz).unit.is_equivalent(units.K)


def test_Trcv_func_bad(bm):
    with pytest.raises(ValueError, match="Trcv function must return a temperature"):
        Observatory(antpos=np.zeros((3, 3)) * units.m, beam=bm, Trcv=lambda f: 3)


def test_observatory(bm):
    a = Observatory(antpos=np.zeros((3, 3)) * units.m, beam=bm)
    assert a.frequency == bm.frequency
    assert a.baselines_metres.shape == (3, 3, 3)
    assert (a.baselines_metres * a.metres_to_wavelengths).unit == units.dimensionless_unscaled
    assert a.baseline_lengths.shape == (3, 3)
    assert np.all(a.baseline_lengths == 0)

    b = Observatory(antpos=np.array([[0, 0, 0], [1, 0, 0], [3, 0, 0]]) * units.m, beam=bm)
    assert units.isclose(b.shortest_baseline / b.metres_to_wavelengths, 1 * units.m, rtol=1e-3)
    assert units.isclose(b.longest_baseline / b.metres_to_wavelengths, 3 * units.m, rtol=1e-3)
    assert b.observation_duration < 1 * units.day
    assert len(b.get_redundant_baselines()) == 6  # including swapped ones
    with pytest.raises(AssertionError):
        b.time_offsets_from_obs_int_time(b.observation_duration * 1.1)

    assert len(b.time_offsets_from_obs_int_time(b.observation_duration / 1.05)) == 2
    assert units.isclose(
        b.longest_used_baseline() / b.metres_to_wavelengths, 3 * units.m, rtol=1e-3
    )


def test_grid_baselines(bm):
    rng = np.random.default_rng(1234)
    a = Observatory(antpos=rng.normal(loc=0, scale=50, size=(20, 3)) * units.m, beam=bm)
    bl_groups = a.get_redundant_baselines()
    bl_coords = a.baseline_coords_from_groups(bl_groups)
    bl_counts = a.baseline_weights_from_groups(bl_groups)

    grid0 = a.grid_baselines(coherent=True)
    grid1 = a.grid_baselines(coherent=True, baselines=bl_coords, weights=bl_counts)
    assert np.allclose(grid0, grid1)


def test_min_max_antpos(bm):
    a = Observatory(
        antpos=np.array([np.linspace(0, 50, 11), np.zeros(11), np.zeros(11)]).T * units.m,
        beam=bm,
        min_antpos=7 * units.m,
    )

    assert len(a.antpos) == 9

    a = Observatory(
        antpos=np.array([np.linspace(0, 50, 11), np.zeros(11), np.zeros(11)]).T * units.m,
        beam=bm,
        max_antpos=10 * units.m,
    )

    assert len(a.antpos) == 2


def test_from_uvdata(bm):
    uv = pyuvdata.UVData()
    uv.telescope.antenna_positions = (
        np.array([[0, 0, 0], [0, 1, 0], [1, 0, 0], [40, 0, 40]]) * units.m
    )
    uv.telescope.location = [x.value for x in EarthLocation.from_geodetic(0, 0).to_geocentric()]
    uv.telescope.antenna_positions = (
        np.array([[0, 0, 0], [0, 1, 0], [1, 0, 0], [40, 0, 40]]) * units.m
    )
    uv.telescope.location = EarthLocation.from_geodetic(0, 0)

    a = Observatory.from_uvdata(uvdata=uv, beam=bm)
    assert np.all(a.antpos == uv.telescope.antenna_positions)


def test_different_antpos_loaders(tmp_path: Path):
    antpos = np.array([[0, 0, 0], [14, 0, 0], [28, 0, 0], [70, 0, 0]])

    np.save(tmp_path / "antpos.npy", antpos)
    np.savetxt(tmp_path / "antpos.txt", antpos)

    beamtxt = """
    beam:
        class: GaussianBeam
        frequency: !astropy.units.Quantity
            unit: !astropy.units.Unit {unit: MHz}
            value: 150
        dish_size: !astropy.units.Quantity
            unit: !astropy.units.Unit {unit: m}
            value: 14.0
    """

    yamlnpy = f"""
    antpos: !astropy.units.Quantity
        unit: !astropy.units.Unit {{unit: m}}
        value: !npy {tmp_path}/antpos.npy
    {beamtxt}
    """

    with open(tmp_path / "npy.yml", "w") as fl:
        fl.write(yamlnpy)

    obsnpy = Observatory.from_yaml(tmp_path / "npy.yml")

    yamltxt = f"""
    antpos: !astropy.units.Quantity
        unit: !astropy.units.Unit {{unit: m}}
        value: !txt {tmp_path}/antpos.txt
    {beamtxt}
    """
    with open(tmp_path / "txt.yml", "w") as fl:
        fl.write(yamltxt)

    obstxt = Observatory.from_yaml(tmp_path / "txt.yml")

    assert obsnpy == obstxt


def test_longest_used_baseline(bm):
    a = Observatory(antpos=np.array([[0, 0, 0], [1, 0, 0], [2, 0, 0]]) * units.m, beam=bm)

    assert np.isclose(a.longest_used_baseline() / a.metres_to_wavelengths, 2 * units.m, atol=1e-3)
    assert np.isclose(
        a.longest_used_baseline(bl_max=1.5 * units.m) / a.metres_to_wavelengths,
        1 * units.m,
        atol=1e-4,
    )


def test_from_yaml(bm):
    rng = np.random.default_rng(1234)
    obs = Observatory.from_yaml(
        {
            "antpos": rng.random((20, 3)) * units.m,
            "beam": {
                "class": "GaussianBeam",
                "frequency": 150 * units.MHz,
                "dish_size": 14 * units.m,
            },
        }
    )
    assert obs.beam == bm

    with pytest.raises(ValueError, match="yaml_file must be a string filepath"):
        Observatory.from_yaml(3)


def test_from_ska():
    pytest.importorskip("ska_ost_array_config")

    from ska_ost_array_config import UVW
    from ska_ost_array_config.array_config import LowSubArray
    from ska_ost_array_config.simulation_utils import simulate_observation

    obs = Observatory.from_ska(subarray_type="AA*", array_type="low", frequency=300.0 * units.MHz)
    low_aastar = LowSubArray(subarray_type="AA*")
    assert obs.antpos.shape == low_aastar.array_config.xyz.data.shape
    Observatory.from_ska(subarray_type="AA*", array_type="mid", frequency=300.0 * units.MHz)
    obs = Observatory.from_ska(subarray_type="AA4", array_type="low", frequency=300.0 * units.MHz)
    low_aa4 = LowSubArray(subarray_type="AA4")
    assert obs.antpos.shape == low_aa4.array_config.xyz.data.shape
    obs = Observatory.from_ska(
        subarray_type="custom",
        array_type="low",
        Trcv=100.0 * units.K,
        frequency=150.0 * units.MHz,
        custom_stations="C*,E1-*",
        exclude_stations="C1,C2",
    )
    low_custom = LowSubArray(
        subarray_type="custom", custom_stations="C*,E1-*", exclude_stations="C1,C2"
    )  # selects all core stations, 6 stations in the E1 cluster, excludes core stations C1 and C2
    assert obs.antpos.shape == low_custom.array_config.xyz.data.shape

    # Simulate visibilities and retreive the UVW values
    ref_time = Time.now()
    zenith = SkyCoord(
        alt=90 * units.deg,
        az=0 * units.deg,
        frame="altaz",
        obstime=ref_time,
        location=low_custom.array_config.location,
    ).icrs
    vis = simulate_observation(
        array_config=low_custom.array_config,
        phase_centre=zenith,
        start_time=ref_time,
        ref_freq=50e6,  # Dummy value. We are after uvw values in [m]
        chan_width=1e3,  # Dummy value. We are after uvw values in [m]
        n_chan=1,
    )
    uvw = UVW.UVW(vis, ignore_autocorr=False)
    uvw_m = uvw.uvdist_m
    assert np.allclose(obs.longest_baseline / obs.metres_to_wavelengths, uvw_m.max() * units.m)

    with pytest.raises(ValueError, match="array_type must be"):
        Observatory.from_ska(
            subarray_type="AA*", array_type="non-existent", frequency=300.0 * units.MHz
        )


def test_get_redundant_baselines(bm):
    a = Observatory(antpos=np.array([[0, 0, 0], [1, 0, 0], [2, 0, 0]]) * units.m, beam=bm)

    reds = a.get_redundant_baselines()
    assert len(reds) == 4  # len-1, len-2 and backwards

    reds = a.get_redundant_baselines(baseline_filters=BaselineRange(bl_max=1.5 * units.m))
    assert len(reds) == 2  # len-1


def test_no_up_coordinate(tmp_path: Path):
    mwafl = PATH / "antpos" / "mwa_phase2_compact_antpos.txt"
    enu = np.genfromtxt(mwafl)

    # Save with only EN coordinates
    with open(tmp_path / "mwa_antpos.txt", "w") as fl:
        np.savetxt(fl, enu[:, :2])

    new_yaml = f"""
antpos: !astropy.units.Quantity
  value: !txt "{tmp_path}/mwa_antpos.txt"
  unit: !astropy.units.Unit {{unit: m}}
beam:
  class: GaussianBeam
  frequency: !astropy.units.Quantity
    unit: !astropy.units.Unit {{unit: MHz}}
    value: 150
  dish_size: !astropy.units.Quantity
    unit: !astropy.units.Unit {{unit: m}}
    value: 35
latitude: !astropy.units.Quantity
  unit: !astropy.units.Unit {{unit: rad}}
  value: -0.4681819
Trcv: !astropy.units.Quantity
  unit: !astropy.units.Unit {{unit: K}}
  value: 100
"""

    with open(tmp_path / "mwa.yaml", "w") as fl:
        fl.write(new_yaml)

    obs = Observatory.from_yaml(tmp_path / "mwa.yaml")
    assert np.all(obs.antpos[:, 2] == 0)


def test_setting_freq_in_profile():
    obs = Observatory.from_profile("MWA-PhaseII", frequency=75 * units.MHz)
    assert obs.frequency == 75 * units.MHz



"""Utility functions for 21cmSense."""

import numpy as np
from astropy import units as un
from astropy.coordinates import EarthLocation, SkyCoord
from astropy.time import Time
from lunarsky import MoonLocation
from lunarsky import SkyCoord as LunarSkyCoord
from lunarsky import Time as LTime
from pyuvdata import utils as uvutils


def between(xmin, xmax):
    """Return an attrs validation function that checks a number is within bounds."""

    def validator(instance, att, val):
        assert xmin <= val <= xmax

    return validator


def positive(instance, att, x):
    """Check that a value is positive.

    This is an attrs validator.
    """
    assert x > 0, "must be positive"


def nonnegative(instance, att, x):
    """Check that a value is non-negative.

    This is an attrs validator.
    """
    assert x >= 0, "must be non-negative"


def find_nearest(array, value):
    """Find closest value in `array` to `value`."""
    return np.abs(array.reshape(-1, 1) - value).argmin(0)


@un.quantity_input
def phase_past_zenith(
    time_past_zenith: un.hour,
    bls_enu: np.ndarray,
    latitude: float,
    world: str = "earth",
    phase_center_dec: un.rad = None,
    use_apparent: bool = True,
):
    """Compute UVWs phased to a point rotated from zenith by a certain amount of time.

    This function specifies a longitude and time of observation without loss of
    generality -- all that matters is the time since a hypothetical point was at zenith,
    and the latitude of the array.

    Parameters
    ----------
    time_past_zenith
        The time passed since the point was at its closest to zenith. Must be a
        quantity with time units.
    bls_enu
        An (Nbls, 3)-shaped array containing the ENU coordinates of the baseline
        vectors (equivalent to the UVWs if phased to zenith).
    latitude
        The latitude of the center of the array, in radians.
    world
        Whether the telescope is on the Earth or Moon.
    phase_center_dec
        If given, the declination of the phase center. If not given, it is set to the
        latitude of the array (i.e. the phase center passes through zenith).
    use_apparent
        Whether to use the apparent coordinates of the phase center (i.e. after
        accounting for nutation and precession etc.)

    Returns
    -------
    uvws
        The array of UVWs correctly phased.
    """
    # Generate ra/dec of zenith at time in the phase_frame coordinate system
    # to use for phasing
    if world == "earth":
        telescope_location = EarthLocation.from_geodetic(lon=0, lat=latitude)
    else:
        telescope_location = MoonLocation.from_selenodetic(lon=0, lat=latitude)

    # JD is arbitrary
    jd = 2454600

    if world == "earth":
        tm = Time(jd, format="jd", location=telescope_location)

        phase_center_coord = SkyCoord(
            alt=90 * un.deg,
            az=0 * un.deg,
            obstime=tm,
            frame="altaz",
            location=telescope_location,
        )
    else:
        tm = LTime(jd, format="jd", location=telescope_location)

        phase_center_coord = LunarSkyCoord(
            alt=90 * un.deg,
            az=0 * un.deg,
            obstime=tm,
            frame="lunartopo",
            location=telescope_location,
        )

    phase_center_coord = phase_center_coord.transform_to("icrs")

    if phase_center_dec is not None:
        phase_center_coord = phase_center_coord.__class__(
            ra=phase_center_coord.ra,
            dec=phase_center_dec,
            obstime=tm,
            frame=phase_center_coord.frame,
            location=telescope_location,
        )

    obstimes = phase_center_coord.obstime + time_past_zenith
    lsts = obstimes.sidereal_time("apparent", longitude=0.0).rad

    if not hasattr(lsts, "__len__"):
        lsts = np.array([lsts])

    if use_apparent:
        app_ra, app_dec = uvutils.phasing.calc_app_coords(
            lon_coord=phase_center_coord.ra.to_value("rad"),
            lat_coord=phase_center_coord.dec.to_value("rad"),
            time_array=obstimes.utc.jd,
            telescope_loc=telescope_location,
        )

        app_ra = np.tile(app_ra, len(bls_enu))
        app_dec = np.tile(app_dec, len(bls_enu))

    else:
        app_ra = phase_center_coord.ra.to_value("rad") * np.ones(len(bls_enu) * len(lsts))
        app_dec = phase_center_coord.dec.to_value("rad") * np.ones(len(bls_enu) * len(lsts))

    # Now make everything nbls * ntimes big.
    _lsts = np.tile(lsts, len(bls_enu))
    uvws = np.repeat(bls_enu, len(lsts), axis=0)

    out = uvutils.phasing.calc_uvw(
        app_ra=app_ra,
        app_dec=app_dec,
        lst_array=_lsts,
        uvw_array=uvws,
        telescope_lat=latitude.to_value("rad"),
        telescope_lon=0.0,
        from_enu=True,
        use_ant_pos=False,
    )
    return out.reshape((len(bls_enu), len(lsts), 3))



"""Test the I/O module."""

import hickle
import numpy as np
from astropy import units as un

from py21cmsense import GaussianBeam, Observation, Observatory

rng = np.random.default_rng(1234)
beam = GaussianBeam(frequency=150 * un.MHz, dish_size=14 * un.m)
obs = Observatory(beam=beam, antpos=rng.random((25, 3)) * 30 * un.m)
observation = Observation(observatory=obs)


def test_beam(tmpdirec):
    hickle.dump(beam, tmpdirec / "beam.h5")
    new = hickle.load(tmpdirec / "beam.h5")

    assert new == beam


def test_observatory(tmpdirec):
    hickle.dump(obs, tmpdirec / "observatory.h5")
    new = hickle.load(tmpdirec / "observatory.h5")
    assert new == obs


def test_observation(tmpdirec):
    hickle.dump(observation, tmpdirec / "observation.h5")
    new = hickle.load(tmpdirec / "observation.h5")
    assert new == observation



"""Tests of the phasing code for calculating UVWs."""

import numpy as np
import pytest
from astropy import units as un
from astropy.coordinates import EarthLocation, SkyCoord
from astropy.time import Time
from pyuvdata import utils as uvutils

from py21cmsense._utils import phase_past_zenith


@pytest.mark.parametrize("lat", [-1.0, -0.5, 0, 0.5, 1.0])
@pytest.mark.parametrize("use_apparent", [True, False])
def test_phase_at_zenith(lat, use_apparent):
    bls_enu = np.array(
        [
            [1, 0, 0],
            [0, 1, 0],
        ]
    )

    uvws = phase_past_zenith(
        time_past_zenith=0.0 * un.day,
        bls_enu=bls_enu,
        latitude=lat * un.rad,
        world="earth",
        use_apparent=use_apparent,
    )

    assert np.allclose(np.squeeze(uvws), bls_enu, atol=5e-3)


@pytest.mark.parametrize("use_apparent", [True, False])
def test_phase_past_zenith(use_apparent):
    bls_enu = np.array(
        [
            [1, 0, 0],
            [0, 1, 0],
        ]
    )

    # Almost rotated to the horizon.
    uvws = np.squeeze(
        phase_past_zenith(
            time_past_zenith=0.2 * un.day,
            bls_enu=bls_enu,
            latitude=0 * un.rad,
            world="earth",
            use_apparent=use_apparent,
        )
    )

    assert uvws[0][0] < 0.35  # Much foreshortened
    assert np.isclose(uvws[1][1], 1)  # N-S direction doesn't get foreshortened.


def test_phase_past_zenith_shape():
    bls_enu = np.array(
        [
            [1, 0, 0],
            [0, 1, 0],
            [1, 0, 0],
            [0, 10, 0],
            [10, 0, 0],
        ]
    )

    times = np.array([0, 0.1, 0, 0.1]) * un.day

    # Almost rotated to the horizon.
    uvws = phase_past_zenith(
        time_past_zenith=times, bls_enu=bls_enu, latitude=0 * un.rad, world="earth"
    )

    assert uvws.shape == (5, 4, 3)
    assert np.allclose(uvws[0], uvws[2])  # Same baselines
    assert np.allclose(uvws[:, 0], uvws[:, 2])  # Same times
    assert np.allclose(uvws[:, 1], uvws[:, 3])  # Same times


@pytest.mark.parametrize("lat", [-1.0, -0.5, 0, 0.5, 1.0])
def test_use_apparent(lat):
    bls_enu = np.array(
        [
            [1, 0, 0],
            [0, 1, 0],
        ]
    )

    times = np.linspace(-1, 1, 3) * un.hour

    # Almost rotated to the horizon.
    uvws = phase_past_zenith(
        time_past_zenith=times, bls_enu=bls_enu, latitude=lat * un.rad, world="earth"
    )
    uvws0 = phase_past_zenith(
        time_past_zenith=times,
        bls_enu=bls_enu,
        latitude=lat * un.rad,
        world="earth",
        use_apparent=True,
    )

    np.testing.assert_allclose(uvws, uvws0, atol=1e-2)


@pytest.mark.parametrize("lat", [-1.0, -0.5, 0, 0.5, 1.0])
@pytest.mark.parametrize("time_past_zenith", [-1 * un.hour, 0 * un.hour, 1 * un.hour])
def test_calc_app_coords(lat, time_past_zenith):
    # Generate ra/dec of zenith at time in the phase_frame coordinate system
    # to use for phasing
    telescope_location = EarthLocation.from_geodetic(lon=0, lat=lat * un.rad)

    # JD is arbitrary
    jd = 2454600

    zenith_coord = SkyCoord(
        alt=90 * un.deg,
        az=0 * un.deg,
        obstime=Time(jd, format="jd"),
        frame="altaz",
        location=telescope_location,
    )
    zenith_coord = zenith_coord.transform_to("icrs")

    obstime = zenith_coord.obstime + time_past_zenith

    ra = zenith_coord.ra.to_value("rad")
    dec = zenith_coord.dec.to_value("rad")
    app_ra, app_dec = uvutils.phasing.calc_app_coords(
        lon_coord=ra, lat_coord=dec, time_array=obstime.utc.jd, telescope_loc=telescope_location
    )

    assert np.isclose(app_ra, ra, atol=0.02)  # give it 1 degree wiggle room.
    assert np.isclose(app_dec, dec, atol=0.02)



"""A module defining baseline filters.

While you can simply use a function that takes a single baseline (with three
co-ordinates) and returns a bool, this module provides standard kinds of filters (eg.
using baselines within a certain length range). It also enables loading the filters
from string names, useful for YAML files.
"""

import abc

import attr
import numpy as np
from astropy import units as un

from . import units as tp

_ALL_BASELINE_FILTERS = {}


class BaselineFilter(abc.ABC):
    """Abstract base class for theory models.

    Subclasses must implement the :meth:`delta_squared` method.
    """

    def __init_subclass__(cls) -> None:
        """Add the subclass to the plugin dict."""
        _ALL_BASELINE_FILTERS[cls.__name__] = cls
        return super().__init_subclass__()

    @abc.abstractmethod
    def __call__(self, bl: tp.Length) -> bool:
        """Determine whether a baseline should be included or not.

        Parameters
        ----------
        bl
            The baseline, which will be a length-3 array whose elements are
            quantities (in length units, typically metres).

        Returns
        -------
        bool
            True if the baseline should be included.
        """
        # pragma: no cover


@attr.define
class BaselineRange(BaselineFilter):
    """Theory model from EOS2021 (https://arxiv.org/abs/2110.13919)."""

    bl_min: tp.Length = attr.field(default=0 * un.m, validator=tp.vld_physical_type("length"))
    bl_max: tp.Length = attr.field(default=np.inf * un.m, validator=tp.vld_physical_type("length"))
    direction: str = attr.field(default="mag", validator=attr.validators.in_(("ew", "ns", "mag")))

    @bl_max.validator
    def _bl_max_vld(self, att, val):
        if val <= self.bl_min:
            raise ValueError(
                f"bl_max must be greater than bl_min, got bl_min={self.bl_min} and bl_max={val}"
            )

    def __call__(self, bl: tp.Length) -> bool:
        """Determine whether a baseline should be included or not.

        Parameters
        ----------
        bl
            The baseline, which will be a length-3 array whose elements are
            quantities (in length units, typically metres).

        Returns
        -------
        bool
            True if the baseline should be included.
        """
        if self.direction == "ew":
            return self.bl_min <= bl[0] < self.bl_max
        elif self.direction == "ns":
            return self.bl_min <= bl[1] < self.bl_max
        elif self.direction == "mag":
            blsize = np.sqrt(bl[0] ** 2 + bl[1] ** 2)
            return self.bl_min <= blsize < self.bl_max



"""Tests for the Observation class."""

import copy
import pickle

import numpy as np
import pytest
from astropy import units
from astropy.cosmology.units import littleh

from py21cmsense import GaussianBeam, Observation, Observatory


@pytest.fixture(scope="module")
def bm():
    return GaussianBeam(150.0 * units.MHz, dish_size=14 * units.m)


@pytest.fixture(scope="module", params=["earth", "moon"])
def wd(request):
    return request.param


@pytest.fixture(scope="module")
def observatory(bm, wd):
    return Observatory(
        antpos=np.array([[0, 0, 0], [14, 0, 0], [28, 0, 0], [70, 0, 0], [0, 14, 0], [23, -45, 0]])
        * units.m,
        latitude=-32 * units.deg,
        beam=bm,
        world=wd,
    )


def test_units(observatory):
    obs = Observation(observatory=observatory)

    assert obs.time_per_day.unit == units.hour
    assert obs.lst_bin_size.to("min").unit == units.min
    assert obs.integration_time.to("s").unit == units.s
    assert obs.bandwidth.to("MHz").unit == units.MHz
    assert obs.bl_min.to("m").unit == units.m
    assert obs.bl_max.to("m").unit == units.m
    assert obs.tsky_amplitude.to("mK").unit == units.mK
    assert obs.tsky_ref_freq.to("MHz").unit == units.MHz

    assert obs.frequency == observatory.frequency
    assert obs.n_lst_bins > 1
    assert obs.Tsky.to("mK").unit == units.mK
    assert obs.Tsys.to("mK").unit == units.mK
    assert obs.Trms.to("mK").unit == units.mK
    assert 6 < obs.redshift < 12
    assert obs.kparallel.unit == littleh / units.Mpc
    assert obs.total_integration_time.to("s").unit == units.s
    assert len(obs.ugrid_edges) == len(obs.ugrid) + 1
    assert obs.clone() == obs


def test_pickle(observatory):
    obs = Observation(observatory=observatory)

    string_rep = pickle.dumps(obs)
    obs2 = pickle.loads(string_rep)
    assert obs == obs2


def test_uvcov(observatory):
    coherent_obs = Observation(observatory=observatory, coherent=True)

    incoherent_obs = Observation(observatory=observatory, coherent=False)

    assert np.all(coherent_obs.uv_coverage >= incoherent_obs.uv_coverage)


def test_equality(observatory):
    new_observatory = copy.deepcopy(observatory)
    assert new_observatory == observatory


def test_from_yaml(observatory):
    rng = np.random.default_rng(1234)
    obs = Observation.from_yaml(
        {
            "observatory": {
                "antpos": rng.random((20, 3)) * units.m,
                "beam": {
                    "class": "GaussianBeam",
                    "frequency": 150 * units.MHz,
                    "dish_size": 14 * units.m,
                },
            }
        }
    )
    assert obs.observatory.antpos.shape == (20, 3)

    with pytest.raises(ValueError, match="yaml_file must be a string filepath"):
        Observation.from_yaml(3)


def test_huge_time_per_day_size(observatory: Observatory, wd):
    tpd = 25 * units.hour if wd == "earth" else 682.5 * units.hour
    with pytest.raises(ValueError, match="time_per_day should be between 0 and"):
        Observation(observatory=observatory, time_per_day=tpd)


def test_huge_track_size(observatory: Observatory, wd):
    tck = 25 * units.hour if wd == "earth" else 682.5 * units.hour
    with pytest.raises(ValueError, match="track should be between 0 and"):
        Observation(observatory=observatory, track=tck)


def test_huge_lst_bin_size(observatory: Observatory, wd):
    lst = 23 * units.hour if wd == "earth" else 627.9 * units.hour
    with pytest.raises(ValueError, match="lst_bin_size must be <= time_per_day"):
        Observation(observatory=observatory, lst_bin_size=lst)

    lst2 = 25 * units.hour if wd == "earth" else 682.5 * units.hour
    with pytest.raises(ValueError, match="lst_bin_size should be between 0 and"):
        Observation(observatory=observatory, lst_bin_size=lst2)


def test_huge_integration_time(observatory: Observatory):
    with pytest.raises(ValueError, match="integration_time must be <= lst_bin_size"):
        Observation(
            observatory=observatory,
            lst_bin_size=1 * units.hour,
            integration_time=2 * units.hour,
        )


def test_trcv_func(observatory: Observatory):
    observatory = observatory.clone(Trcv=lambda f: (f / units.MHz) * 10 * units.mK)

    obs = Observation(
        observatory=observatory,
    )
    assert obs.Trcv.unit == units.K
    assert obs.Trcv == (observatory.beam.frequency / units.MHz) * 0.01 * units.K


def test_non_zenith_pointing(bm):
    """Test that not pointing at zenith gives different results."""
    observatory = Observatory(
        antpos=np.array([[0, 0, 0], [14, 0, 0], [28, 0, 0], [70, 0, 0], [0, 14, 0], [23, -45, 0]])
        * units.m,
        latitude=-32 * units.deg,
        beam=bm,
    )
    at_zenith = Observation(observatory=observatory)

    almost_zenith = Observation(
        observatory=observatory, phase_center_dec=observatory.latitude * 0.99
    )
    np.testing.assert_allclose(at_zenith.uv_coverage, almost_zenith.uv_coverage)

    not_zenith = Observation(
        observatory=observatory,
        phase_center_dec=observatory.latitude + 45 * units.deg,
    )
    assert not np.allclose(not_zenith.uv_coverage, at_zenith.uv_coverage)


def test_non_zenith_pointing_only_ew(bm):
    """Test that not pointing at zenith gives the SAME results if all baselines are EW."""
    observatory = Observatory(
        antpos=np.array([[0, 0, 0], [14, 0, 0], [28, 0, 0], [70, 0, 0]]) * units.m,
        latitude=-32 * units.deg,
        beam=bm,
    )
    at_zenith = Observation(observatory=observatory)

    not_zenith = Observation(
        observatory=observatory,
        phase_center_dec=observatory.latitude + 45 * units.deg,
    )
    np.testing.assert_allclose(not_zenith.uv_coverage, at_zenith.uv_coverage)



"""Test the CLI module."""

import traceback
from os import path

import pytest
from astropy.io.misc import yaml
from click.testing import CliRunner

from py21cmsense import cli

here = path.dirname(path.abspath(__file__))
example_configs = path.join(here, "../example_configs")


@pytest.fixture(scope="module")
def runner():
    return CliRunner()


@pytest.fixture(scope="module")
def observatory_config() -> str:
    return path.join(example_configs, "observatory_hera.yml")


@pytest.fixture(scope="module")
def observation_config(tmpdirec, observatory_config):
    with open(path.join(example_configs, "observation_hera.yml")) as fl:
        observation = yaml.load(fl)

    observation["observatory"] = observatory_config

    with open(path.join(tmpdirec, "observation.yml"), "w") as fl:
        yaml.dump(observation, fl)

    return path.join(tmpdirec, "observation.yml")


@pytest.fixture(scope="module")
def sensitivity_config(tmpdirec, observation_config):
    with open(path.join(example_configs, "sensitivity_hera.yml")) as fl:
        sensitivity = yaml.load(fl)

    sensitivity["observation"] = observation_config

    with open(path.join(tmpdirec, "sensitivity.yml"), "w") as fl:
        yaml.dump(sensitivity, fl)

    return path.join(tmpdirec, "sensitivity.yml")


def test_gridding_baselines(runner, observation_config, tmpdirec):
    output = runner.invoke(
        cli.main, ["grid-baselines", observation_config, "--direc", str(tmpdirec)]
    )
    if output.exception:
        traceback.print_exception(*output.exc_info)

    assert output.exit_code == 0


def test_calc_sense(runner, sensitivity_config, tmpdirec):
    output = runner.invoke(cli.main, ["calc-sense", sensitivity_config, "--direc", str(tmpdirec)])
    if output.exception:
        traceback.print_exception(*output.exc_info)

    assert output.exit_code == 0


def test_both(runner, tmpdirec, observation_config, sensitivity_config):
    output = runner.invoke(
        cli.main,
        [
            "grid-baselines",
            observation_config,
            "--direc",
            tmpdirec,
            "--outfile",
            "arrayfile.h5",
        ],
    )

    if output.exception:
        traceback.print_exception(*output.exc_info)

    output = runner.invoke(
        cli.main,
        [
            "calc-sense",
            sensitivity_config,
            "--array-file",
            path.join(tmpdirec, "arrayfile.h5"),
            "--plot-title",
            "MYTITLE",
        ],
    )
    if output.exception:
        traceback.print_exception(*output.exc_info)

    assert output.exit_code == 0



"""Test the sensitivity module."""

import warnings
from functools import wraps

import numpy as np
import pytest
from astropy import units
from astropy.cosmology.units import littleh
from spiceypy.utils.exceptions import SpiceUNKNOWNFRAME

from py21cmsense import GaussianBeam, Observation, Observatory, PowerSpectrum, theory
from py21cmsense.sensitivity import Sensitivity


def skip_on(exception, reason="Default reason"):
    # Func below is the real decorator and will receive the test function as param
    def decorator_func(f):
        @wraps(f)
        def wrapper(*args, **kwargs):
            try:
                # Try to run the test
                return f(*args, **kwargs)
            except exception:
                # If exception of given type happens
                # just swallow it and raise pytest.Skip with given reason
                pytest.skip(reason)

        return wrapper

    return decorator_func


@pytest.fixture(scope="module")
def bm():
    return GaussianBeam(150.0 * units.MHz, dish_size=14 * units.m)


@pytest.fixture(scope="module", params=["earth", "moon"])
def wd(request):
    return request.param


@pytest.fixture(scope="module")
def observatory(bm, wd):
    return Observatory(
        antpos=np.array([[0, 0, 0], [14, 0, 0], [28, 0, 0], [70, 0, 0]]) * units.m,
        beam=bm,
        world=wd,
    )


@pytest.fixture(scope="module")
def observation(observatory):
    return Observation(observatory=observatory)


def test_units(observation):
    ps = PowerSpectrum(observation=observation)

    assert ps.horizon_buffer.to("littleh/Mpc").unit == littleh / units.Mpc
    assert ps.k1d.to("littleh/Mpc").unit == littleh / units.Mpc
    assert isinstance(ps.power_normalisation(0.1 * littleh / units.Mpc), float)
    assert ps.horizon_limit(10).to("littleh/Mpc").unit == littleh / units.Mpc


@skip_on(SpiceUNKNOWNFRAME, "Unknown FRAME (flaky exception)")
def test_sensitivity_2d(observation):
    ps = PowerSpectrum(observation=observation)
    sense_thermal = ps.calculate_sensitivity_2d(thermal=True, sample=False)
    sense_full = ps.calculate_sensitivity_2d()
    assert all(np.all(sense_thermal[key] <= sense_full[key]) for key in sense_thermal)

    with pytest.raises(ValueError, match="Either thermal or sample must be True"):
        ps.calculate_sensitivity_2d(thermal=False, sample=False)


@skip_on(SpiceUNKNOWNFRAME, "Unknown FRAME (flaky exception)")
def test_sensitivity_2d_grid(observation, caplog):
    ps = PowerSpectrum(observation=observation)
    sense_ungridded = ps.calculate_sensitivity_2d(thermal=True, sample=True)
    kperp = np.array([x.value for x in sense_ungridded]) * next(iter(sense_ungridded.keys())).unit
    sense = ps.calculate_sensitivity_2d_grid(
        kperp_edges=np.linspace(kperp.min().value, kperp.max().value, 10) * kperp.unit,
        kpar_edges=ps.k1d,
    )
    assert sense.shape == (9, len(ps.k1d) - 1)


@skip_on(SpiceUNKNOWNFRAME, "Unknown FRAME (flaky exception)")
def test_sensitivity_1d_binned(observation):
    ps = PowerSpectrum(observation=observation)
    assert np.all(ps.calculate_sensitivity_1d() == ps.calculate_sensitivity_1d_binned(ps.k1d))
    kbins = np.linspace(0.1, 0.5, 10) * littleh / units.Mpc
    sense1d_sample = ps.calculate_sensitivity_1d_binned(k=kbins)
    assert len(sense1d_sample) == len(kbins)


@skip_on(SpiceUNKNOWNFRAME, "Unknown FRAME (flaky exception)")
def test_plots(observation):
    # this is a dumb test, just checking that it doesn't error.
    ps = PowerSpectrum(observation=observation)
    sense2d = ps.calculate_sensitivity_2d()
    ps.plot_sense_2d(sense2d)


@skip_on(SpiceUNKNOWNFRAME, "Unknown FRAME (flaky exception)")
def test_sensitivity_optimistic(observation):
    ps = PowerSpectrum(observation=observation, foreground_model="optimistic")
    assert ps.horizon_limit(10.0) > ps.horizon_limit(5.0)


@skip_on(SpiceUNKNOWNFRAME, "Unknown FRAME (flaky exception)")
def test_sensitivity_foreground_free(observation):
    ps = PowerSpectrum(observation=observation, foreground_model="foreground_free")
    assert ps.horizon_limit(10.0) == 0


@skip_on(SpiceUNKNOWNFRAME, "Unknown FRAME (flaky exception)")
def test_infs_in_trms(observation):
    # default dumb layout should have lots of infs..
    assert np.any(np.isinf(observation.Trms))
    ps = PowerSpectrum(observation=observation)
    ps.calculate_sensitivity_2d()
    # merely get through the calculations...


def test_write_to_custom_filename(observation, tmp_path):
    out = tmp_path / "outfile.h5"
    with warnings.catch_warnings():
        warnings.simplefilter("ignore")
        ps = PowerSpectrum(observation=observation)
        out2 = ps.write(filename=out)
    assert out2 == out


def test_load_yaml_bad():
    with pytest.raises(
        ValueError,
        match="yaml_file must be a string filepath or a raw dict from such a file",
    ):
        Sensitivity.from_yaml(1)

    rng = np.random.default_rng(1234)
    with pytest.raises(ImportError, match="Could not import"):
        PowerSpectrum.from_yaml(
            {
                "plugins": ["this.is.not.a.module"],
                "observatory": {
                    "antpos": rng.random((20, 3)) * units.m,
                    "beam": {
                        "class": "GaussianBeam",
                        "frequency": 150 * units.MHz,
                        "dish_size": 14 * units.m,
                    },
                },
            }
        )


@skip_on(SpiceUNKNOWNFRAME, "Unknown FRAME (flaky exception)")
def test_systematics_mask(observation):
    ps = PowerSpectrum(
        observation=observation,
        systematics_mask=lambda kperp, kpar: np.zeros(len(kpar), dtype=bool),
    )
    assert len(ps.calculate_sensitivity_2d()) == 0


@skip_on(SpiceUNKNOWNFRAME, "Unknown FRAME (flaky exception)")
def test_track(observatory):
    """Test that setting `track` is the same as setting lst_bin_width."""
    obs1 = Observation(observatory=observatory, lst_bin_size=1 * units.hour)
    obs2 = Observation(observatory=observatory, track=1 * units.hour)

    assert np.all(obs1.uv_coverage == obs2.uv_coverage)


def test_clone(observation):
    ps = PowerSpectrum(
        observation=observation,
    )

    ps2 = ps.clone()
    assert ps2 == ps


@skip_on(SpiceUNKNOWNFRAME, "Unknown FRAME (flaky exception)")
def test_at_freq(observation):
    ps = PowerSpectrum(observation=observation, theory_model=theory.EOS2016Bright())
    ps2 = ps.at_frequency(0.9 * observation.frequency)

    assert ps2.frequency == 0.9 * observation.frequency
    with pytest.warns(UserWarning, match="Extrapolating above the simulated theoretical"):
        assert ps.calculate_significance() != ps2.calculate_significance()


def test_bad_theory(observation):
    with pytest.raises(ValueError, match="The theory_model must be an instance of TheoryModel"):
        PowerSpectrum(observation=observation, theory_model=3)



"""Module defining new YAML tags for py21cmsense."""

import inspect
import pickle
from functools import wraps

import numpy as np
import yaml
from astropy import units as un
from astropy.io.misc.yaml import AstropyLoader

_DATA_LOADERS = {}

_YAML_LOADERS = (yaml.FullLoader, yaml.SafeLoader, yaml.Loader, AstropyLoader)


class LoadError(IOError):
    """Error raised on trying to load data from YAML files."""


def data_loader(tag=None):
    """Convert a function into a YAML tag for loading external datafiles.

    The form of the tag is::

        !<tag> <path_to_data> [| <unit>]
    """

    def inner(fnc):
        _DATA_LOADERS[fnc.__name__] = fnc

        new_tag = tag or fnc.__name__.split("_loader")[0]
        fnc.tag = new_tag

        # ensure it only takes path to data.
        assert len(inspect.signature(fnc).parameters) == 1

        @wraps(fnc)
        def wrapper(data):
            try:
                return fnc(data)
            except OSError:
                raise
            except Exception as e:
                raise LoadError(str(e)) from e

        def yaml_fnc(loader, node):
            args = node.value.split("|")
            if len(args) == 1:
                return wrapper(node.value)
            elif len(args) == 2:
                # Return with a unit
                return wrapper(args[0].strip()) * getattr(un, args[1].strip())
            else:
                raise ValueError(
                    f"Too many arguments in {new_tag} tag. "
                    "Should be of the form: "
                    f"!{new_tag} <path_to_data> | <unit>"
                )

        for loader in _YAML_LOADERS:
            yaml.add_constructor(f"!{new_tag}", yaml_fnc, Loader=loader)

        return wrapper

    return inner


@data_loader("pkl")
def pickle_loader(data):
    """YAML tag for loading pickle files."""
    with open(data, "rb") as f:
        data = pickle.load(f)
    return data


@data_loader()
def npz_loader(data):
    """YAML tag for loading npz files."""
    return dict(np.load(data))


@data_loader()
def npy_loader(data):
    """YAML tag for loading npy files."""
    return np.load(data)


@data_loader()
def txt_loader(data):
    """YAML tag for loading ASCII files."""
    return np.genfromtxt(data)


def yaml_func(tag=None):
    """Convert a function into a YAML tag."""

    def inner(fnc):
        new_tag = tag or fnc.__name__
        fnc.tag = new_tag

        def yaml_fnc(loader, node):
            kwargs = loader.construct_mapping(node)
            return fnc(**kwargs)

        for loader in _YAML_LOADERS:
            yaml.add_constructor(f"!{new_tag}", yaml_fnc, Loader=loader)

        return fnc

    return inner



"""Test the conversions module."""

import numpy as np
from astropy import units
from astropy.cosmology import Planck15

from py21cmsense import conversions as cnv


def test_f2z():
    assert np.isclose(cnv.f2z(0.15 * units.GHz), cnv.f2z(150.0 * units.MHz), rtol=1e-6)
    assert np.isclose(cnv.f2z(0.142 * units.GHz), 9, atol=0.1)


def test_z2f():
    assert units.isclose(cnv.z2f(9), 142 * units.MHz, atol=1 * units.MHz)


def test_dL_dth():
    (cnv.dL_dth(10) * units.rad).to("Mpc/littleh")
    cosmo = Planck15.clone(H0=Planck15.H0 / 1.1)
    assert cnv.dL_dth(10, cosmo) < cnv.dL_dth(10)


def test_dl_df():
    (cnv.dL_df(10) * 1 * units.MHz).to("Mpc/littleh")
    cosmo = Planck15.clone(H0=Planck15.H0 / 1.1)
    assert cnv.dL_df(10, cosmo) < cnv.dL_df(10)


def test_dk_du():
    cnv.dk_du(10).to("littleh/Mpc")
    cosmo = Planck15.clone(H0=Planck15.H0 / 1.1)
    assert cnv.dk_du(10, cosmo) > cnv.dk_du(10)


def test_dk_deta():
    (cnv.dk_deta(10) * 1 / units.MHz).to("littleh/Mpc")
    cosmo = Planck15.clone(H0=Planck15.H0 / 1.1)
    assert cnv.dk_deta(10, cosmo) > cnv.dk_deta(10)


def test_X2Y():
    cnv.X2Y(10).to("Mpc^3 / (littleh^3 sr GHz)")
    cosmo = Planck15.clone(H0=Planck15.H0 / 1.1)
    assert cnv.X2Y(10, cosmo) < cnv.X2Y(10)


def test_approx_dL_dth():
    assert np.isclose(cnv.dL_dth(10), cnv.dL_dth(10.0, approximate=True), rtol=0.02)


def test_approx_dL_df():
    assert np.isclose(cnv.dL_df(10), cnv.dL_df(10.0, approximate=True), rtol=0.02)



"""CLI routines for 21cmSense."""

import logging
import tempfile
from os import path
from pathlib import Path

import click
from astropy.io.misc import yaml
from hickle import hickle
from rich.logging import RichHandler

from . import observation
from . import sensitivity as sense

try:
    import matplotlib.pyplot as plt

    HAVE_MPL = True
except ImportError:  # pragma: no cover
    HAVE_MPL = False

main = click.Group()

FORMAT = "%(message)s"
logging.basicConfig(level=logging.INFO, format=FORMAT, datefmt="[%X]", handlers=[RichHandler()])

logger = logging.getLogger("py21cmsense")


@main.command()
@click.argument("configfile", type=click.Path(exists=True, dir_okay=False))
@click.option(
    "--direc",
    type=click.Path(exists=True, dir_okay=True, file_okay=False),
    help="directory to save output file",
    default=".",
)
@click.option(
    "--outfile",
    "-o",
    type=click.Path(exists=False, dir_okay=False, file_okay=True),
    help="filename of output file",
    default=None,
)
def grid_baselines(configfile, direc, outfile):
    """Grid baselines according to CONFIGFILE."""
    obs = observation.Observation.from_yaml(configfile)

    if outfile is None:
        outfile = Path(direc) / (
            f"blmin{obs.bl_min.value:.3f}_blmax{obs.bl_max.value:.3f}_"
            f"{obs.frequency.to('GHz').value:.3f}GHz_observation.h5"
        )
    elif not Path(outfile).is_absolute():
        outfile = Path(direc) / outfile

    hickle.dump(obs, outfile)

    logger.info(f"There are {len(obs.baseline_groups)} baseline types")
    logger.info(f"Saving array file as {outfile}")


@main.command()
@click.argument("configfile", type=click.Path(exists=True, dir_okay=False))
@click.option(
    "--array-file",
    default=None,
    type=click.Path(exists=True, dir_okay=False, file_okay=True),
    help="array file created with grid-baselines",
)
@click.option(
    "--direc",
    type=click.Path(exists=True, dir_okay=True, file_okay=False),
    help="directory to save output file",
    default=".",
)
@click.option("--fname", default=None, type=click.Path(), help="filename to save output file")
@click.option("--thermal/--no-thermal", default=True, help="whether to include thermal noise")
@click.option(
    "--samplevar/--no-samplevar",
    default=True,
    help="whether to include sample variance",
)
@click.option(
    "--write-significance/--no-significance",
    default=True,
    help="whether to write the significance of the PS to screen",
)
@click.option(
    "-p/-P",
    "--plot/--no-plot",
    default=True,
    help="whether to plot the 1D power spectrum uncertainty",
)
@click.option("--plot-title", default=None, type=str, help="title for the output 1D plot")
@click.option("--prefix", default="", type=str, help="string prefix for all output files")
def calc_sense(
    configfile,
    array_file,
    direc,
    fname,
    thermal,
    samplevar,
    write_significance,
    plot,
    plot_title,
    prefix,
):
    """Calculate the sensitivity of an array.

    This is the primary command of 21cmSense, and can be run independently for a
    complete sensitivity calculation.
    """
    if plot and not HAVE_MPL:  # pragma: no cover
        raise click.ClickException(
            "matplotlib is required for plotting, but it is not installed. "
            "Use --no-plot to disable plotting."
        )

    # If given an array-file, overwrite the "observation" parameter
    # in the config with the hickled array file, which has already
    # calculated the uv_coverage.
    if array_file is not None:
        with open(configfile) as fl:
            cfg = yaml.load(fl)
        cfg["observation"] = path.abspath(array_file)

        configfile = tempfile.mktemp()
        with open(configfile, "w") as fl:
            yaml.dump(cfg, fl)

    sensitivity = sense.PowerSpectrum.from_yaml(configfile)
    logger.info(
        f"Used {len(sensitivity.k1d)} bins between "
        f"{sensitivity.k1d.min()} and {sensitivity.k1d.max()}"
    )
    sensitivity.write(filename=fname, thermal=thermal, sample=samplevar, direc=direc, prefix=prefix)

    if write_significance:
        sig = sensitivity.calculate_significance(thermal=thermal, sample=samplevar)
        logger.info(f"Significance of detection: {sig}")

    if plot and HAVE_MPL:
        fig = sensitivity.plot_sense_1d(thermal=thermal, sample=samplevar)
        if plot_title:
            plt.title(plot_title)
        prefix = f"{prefix}_" if prefix else ""
        fig.savefig(
            f"{direc}/{prefix}{sensitivity.foreground_model}_"
            f"{sensitivity.observation.frequency:.3f}.png"
        )



"""Simplistic beam definitions."""

from __future__ import annotations

from abc import ABCMeta, abstractmethod

import attr
from astropy import constants as cnst
from astropy import units as un
from hickleable import hickleable

from . import _utils as ut
from . import units as tp


@hickleable(evaluate_cached_properties=True)
@attr.s
class PrimaryBeam(metaclass=ABCMeta):
    """A Base class defining a Primary Beam and the methods it requires to define.

    Required methods on subclasses are :meth:`area`, :meth:`width`, :meth:`first_null`,
    :meth:`sq_area` and :meth:`uv_resolution`..

    Note that 21cmSense currently only uses the beam width as a means to calculate
    the beam-crossing time, so precise shape is not very important. For that reason,
    it is not very important to implement beam sub-classes.
    """

    frequency: tp.Frequency = attr.ib(
        validator=(tp.vld_physical_type("frequency"), ut.positive),
    )

    def new(self, **kwargs) -> PrimaryBeam:
        """Return a clone of this instance, but change kwargs."""
        return attr.evolve(self, **kwargs)

    def at(self, frequency: tp.Frequency) -> PrimaryBeam:
        """Get a copy of the object at a new frequency."""
        return attr.evolve(self, frequency=frequency)

    @property
    @abstractmethod
    def area(self) -> un.Quantity[un.steradian]:
        """Beam area [units: sr]."""

    @property
    @abstractmethod
    def width(self) -> un.Quantity[un.radians]:
        """Beam width [units: rad]."""

    @property
    @abstractmethod
    def first_null(self) -> un.Quantity[un.radians]:
        """An approximation of the first null of the beam."""

    @property
    @abstractmethod
    def sq_area(self) -> un.Quantity[un.steradian]:
        """The area of the beam^2."""

    @property
    def b_eff(self) -> un.Quantity[un.steradian]:
        r"""Get the effective beam area (Parsons 2014).

        Defined as :math:`(\int B(\Omega) d \Omega)^2 / \int B^2 d\Omega`.
        """
        return self.area**2 / self.sq_area

    @property
    @abstractmethod
    def uv_resolution(self) -> un.Quantity[1 / un.radians]:
        """The UV footprint of the beam."""

    @classmethod
    def from_uvbeam(cls) -> PrimaryBeam:
        """Generate the beam object from a :class:`pyuvdata.UVBeam` object."""
        raise NotImplementedError()


@attr.s(frozen=True)
class GaussianBeam(PrimaryBeam):
    """
    A simple Gaussian Primary beam.

    Parameters
    ----------
    frequency
        The fiducial frequency at which the beam operates, assumed to be in MHz
        unless otherwise defined.
    dish_size
        The size of the (assumed circular) dish, assumed to be in meters unless
        otherwise defined. This generates the beam size.
    """

    dish_size: tp.Length = attr.ib(validator=(tp.vld_physical_type("length"), ut.positive))

    @property
    def wavelength(self) -> un.Quantity[un.m]:
        """The wavelength of the observation."""
        return (cnst.c / self.frequency).to("m")

    @property
    def dish_size_in_lambda(self) -> float:
        """The dish size in units of wavelengths."""
        return (self.dish_size / (cnst.c / self.frequency)).to("").value

    @property
    def uv_resolution(self) -> un.Quantity[1 / un.radian]:
        """The appropriate resolution of a UV cell given the beam size."""
        return self.dish_size_in_lambda

    @property
    def area(self) -> un.Quantity[un.steradian]:
        """The integral of the beam over angle, in sr."""
        return 1.13 * self.fwhm**2

    @property
    def width(self) -> un.Quantity[un.radian]:
        """The width of the beam (i.e. sigma), in radians."""
        return un.rad * 0.45 / self.dish_size_in_lambda

    @property
    def fwhm(self) -> un.Quantity[un.radians]:
        """The full-width half maximum of the beam."""
        return 2.35 * self.width

    @property
    def sq_area(self) -> un.Quantity[un.steradian]:
        """The integral of the squared beam, in sr.

        If frequency is not given, uses the instance's `frequency`
        """
        return self.area / 2

    @property
    def first_null(self) -> un.Quantity[un.radians]:
        """The angle of the first null of the beam.

        .. note:: The Gaussian beam has no null, and in this case we use the first null
                  for an airy disk.
        """
        return un.rad * 1.22 / self.dish_size_in_lambda

    @classmethod
    def from_uvbeam(cls):
        """Construct the beam from a :class:`pyuvdata.UVBeam` object."""
        raise NotImplementedError("Coming Soon!")

    def clone(self, **kwargs):
        """Create a new beam with updated parameters."""
        return attr.evolve(self, **kwargs)



"""Test baseline_filters module."""

import numpy as np
import pytest
from astropy import units as un

from py21cmsense.baseline_filters import BaselineRange

# Test IDs for parametrization
happy_path_ids = [
    "east-west_min",
    "north-south_min",
    "magnitude_min",
    "east-west_max",
    "north-south_max",
    "magnitude_max",
]

edge_case_ids = [
    "east-west_zero",
    "north-south_zero",
    "magnitude_zero",
]

# Happy path test values
happy_path_values = [
    (0 * un.m, np.inf * un.m, "ew", np.array([1, 0, 0]) * un.m, True),
    (0 * un.m, np.inf * un.m, "ns", np.array([0, 1, 0]) * un.m, True),
    (0 * un.m, np.inf * un.m, "mag", np.array([1, 1, 0]) * un.m, True),
    (1 * un.m, np.inf * un.m, "ew", np.array([2, 0, 0]) * un.m, True),
    (1 * un.m, np.inf * un.m, "ns", np.array([0, 2, 0]) * un.m, True),
    (1 * un.m, np.inf * un.m, "mag", np.array([1, 1, 0]) * un.m, True),
]

# Edge case test values
edge_case_values = [
    (0 * un.m, np.inf * un.m, "ew", np.array([0, 0, 0]) * un.m, True),
    (0 * un.m, np.inf * un.m, "ns", np.array([0, 0, 0]) * un.m, True),
    (0 * un.m, np.inf * un.m, "mag", np.array([0, 0, 0]) * un.m, True),
]


@pytest.mark.parametrize(
    ("bl_min", "bl_max", "direction", "baseline", "expected"),
    happy_path_values,
    ids=happy_path_ids,
)
def test_happy_path(bl_min, bl_max, direction, baseline, expected):
    # Arrange
    baseline_range = BaselineRange(bl_min=bl_min, bl_max=bl_max, direction=direction)

    # Act
    result = baseline_range(baseline)

    # Assert
    assert result == expected


@pytest.mark.parametrize(
    ("bl_min", "bl_max", "direction", "baseline", "expected"),
    edge_case_values,
    ids=edge_case_ids,
)
def test_edge_cases(bl_min, bl_max, direction, baseline, expected):
    # Arrange
    baseline_range = BaselineRange(bl_min=bl_min, bl_max=bl_max, direction=direction)

    # Act
    result = baseline_range(baseline)

    # Assert
    assert result == expected


def test_error_cases():
    with pytest.raises(ValueError, match="bl_max must be greater than bl_min"):
        baseline_range = BaselineRange(bl_min=2 * un.m, bl_max=1 * un.m, direction="mag")

    with pytest.raises(ValueError, match="must be in"):
        baseline_range = BaselineRange(bl_min=1 * un.m, bl_max=2 * un.m, direction="invalid")

    baseline_range = BaselineRange(bl_min=1 * un.m, bl_max=2 * un.m, direction="ew")

    with pytest.raises(ValueError, match="Can only apply"):
        baseline_range(np.array([1, 1, 0]))



"""A module defining interferometric observation objects."""

from __future__ import annotations

import collections
from collections import defaultdict
from functools import cached_property
from os import path
from typing import Any, Callable

import attr
import numpy as np
from astropy import units as un
from astropy.cosmology import LambdaCDM, Planck15
from astropy.io.misc import yaml
from attr import validators as vld
from hickleable import hickleable

from . import _utils as ut
from . import conversions as conv
from . import observatory as obs
from . import units as tp


@hickleable(evaluate_cached_properties=True)
@attr.s(kw_only=True)
class Observation:
    """
    A class defining an interferometric Observation.

    Note that to achieve the same behaviour as the original 21cmSense's ``--track``
    option (i.e. track-scan instead of drift-scan) merely requires setting ``lst_bin_size``
    to the length of the tracked scan.

    Parameters
    ----------
    observatory : :class:`~py21cmsense.observatory.Observatory`
        An object defining attributes of the observatory itself (its location etc.)
    hours_per_day : float or Quantity, optional
        The number of good observing hours per day.  This corresponds to the size of a
        low-foreground region in right ascension for a drift scanning instrument on the Earth.
        The total observing time is `n_days*hours_per_day` where a day is 24 hours on the
        Earth and 655.2 hours on the moon.  Default is 6 (163.8 on moon).
        If simulating a tracked scan, `hours_per_day` should be a multiple of the length
        of the track (i.e. for two three-hour tracks per day, `hours_per_day` should be 6).
    track
        Tracked scan observation duration. This is an alias for ``lst_bin_size``, and
        is provided only for ease of use for those familiar with 21cmSense v1.
    lst_bin_size : float or Quantity, optional
        The time assigned to a single LST bin, by default the time it takes for a source
        to travel through the beam's FWHM. If a float, assumed to be in minutes.
    integration_time : float or Quantity, optional
        The amount of time integrated into a single visibility, by default a minute.
        If a float, assumed to be in seconds.
    n_channels : int, optional
        Number of channels used in the observation. Defaults to 82, which is equivalent
        to 1024 channels over 100 MHz of bandwidth. Sets maximum k_parallel that can be
        probed, but little to no overall effect on sensitivity.
    bandwidth : float or Quantity, optional
        The bandwidth used for the observation, assumed to be in MHz. Note this is not the total
        instrument bandwidth, but the redshift range that can be considered co-eval.
    n_days : int, optional
        The number of days observed (for the same set of LSTs). The default is 180 (6 on moon),
        which is the maximum a particular R.A. can be observed in one year if one only observes
        at night. The total observing time is `n_days*hours_per_day` where a day is 24 hours on
        the Earth and 655.2 hours on the moon.
    baseline_filters
        A function that takes a single value: a length-3 array of baseline co-ordinates,
        and returns a bool indicating whether to include the baseline. Built-in filters
        are provided in the :mod:`~baseline_filters` module.
    redundancy_tol : int, optional
        The number of decimal places to which baseline vectors must match (in
        all dimensions) to be considered redundant.
    coherent : bool, optional
        Whether to add different baselines coherently if they are not instantaneously redundant.
    spectral_index : float, optional
        The spectral index of the foreground model. The foreground model is approximated as
        a spatially-independent power-law, and used only for generating sky noise temperature.
        The default is 2.6, based on Mozdzen et al. 2017: 2017MNRAS.464.4995M, figure 8,
        with Galaxy down (see also `tsky_amplitude` and `tsky_ref_freq`).
    tsky_amplitude : float or Quantity, optional
        The temperature of foregrounds at `tsky_ref_freq`. See `spectral_index`.
        Default assumed to be in mK.
    tsky_ref_freq : float or Quantity
        Frequency at which the foreground model is equal to `tsky_amplitude`.
        See `spectral_index`. Default assumed to be in MHz.
    use_approximate_cosmo : bool
        Whether to use approximate cosmological conversion factors. Doing so will give
        the same results as the original 21cmSense code, but non-approximate versions
        that use astropy are preferred.
    cosmo : LambdaCDM
        An astropy cosmology object to use.
    phase_center_dec
        The declination of the phase center (typically used for a tracked scan
        observation). By default, this is set to the latitude of the observatory
        (i.e. the phase-center passes through zenith, like a typical drift-scan
        observation).
    """

    observatory: obs.Observatory = attr.ib(validator=vld.instance_of(obs.Observatory))

    time_per_day: tp.Time = attr.ib(
        validator=(tp.vld_physical_type("time")),
    )
    track: tp.Time | None = attr.ib(
        None,
        validator=attr.validators.optional([tp.vld_physical_type("time")]),
    )
    lst_bin_size: tp.Time = attr.ib(
        validator=(tp.vld_physical_type("time")),
    )
    integration_time: tp.Time = attr.ib(
        60 * un.second, validator=(tp.vld_physical_type("time"), ut.positive)
    )
    n_channels: int = attr.ib(82, converter=int, validator=ut.positive)
    bandwidth: tp.Frequency = attr.ib(
        8 * un.MHz, validator=(tp.vld_physical_type("frequency"), ut.positive)
    )
    n_days: int = attr.ib(converter=int, validator=ut.positive)
    baseline_filters: tuple[Callable[[tp.Length], bool]] = attr.ib(
        default=(), converter=tp._tuplify
    )
    redundancy_tol: int = attr.ib(default=1, converter=int, validator=ut.nonnegative)
    coherent: bool = attr.ib(default=True, converter=bool)

    # The following defaults are based on Mozdzen et al. 2017: 2017MNRAS.464.4995M,
    # figure 8, with galaxy down.
    spectral_index: float = attr.ib(default=2.6, converter=float, validator=ut.between(1.5, 4))
    tsky_amplitude: tp.Temperature = attr.ib(
        default=260000 * un.mK,
        validator=ut.nonnegative,
    )
    tsky_ref_freq: tp.Frequency = attr.ib(default=150 * un.MHz, validator=ut.positive)
    use_approximate_cosmo: bool = attr.ib(default=False, converter=bool)
    cosmo: LambdaCDM = attr.ib(default=Planck15, converter=Planck15.from_format)
    phase_center_dec = attr.ib(validator=(tp.vld_physical_type("angle")))

    @classmethod
    def from_yaml(cls, yaml_file):
        """Construct an :class:`Observation` from a YAML file."""
        if isinstance(yaml_file, str):
            with open(yaml_file) as fl:
                data = yaml.load(fl)
        elif isinstance(yaml_file, collections.abc.Mapping):
            data = yaml_file
        else:
            raise ValueError("yaml_file must be a string filepath or a raw dict from such a file.")

        if (
            isinstance(data["observatory"], str)
            and isinstance(yaml_file, str)
            and not path.isabs(data["observatory"])
        ):
            data["observatory"] = path.join(path.dirname(yaml_file), data["observatory"])

        observatory = obs.Observatory.from_yaml(data.pop("observatory"))
        return cls(observatory=observatory, **data)

    def __gethstate__(self) -> dict[str, Any]:
        """Get the hickle state."""
        d = attr.asdict(self, recurse=False)
        d["cosmo"] = d["cosmo"].to_format("mapping")
        del d["cosmo"]["cosmology"]  # The class.
        return d

    def __sethstate__(self, d: dict[str, Any]) -> None:
        """Set the hickle state."""
        d["cosmo"]["cosmology"] = type(Planck15)
        d["cosmo"] = Planck15.from_format(d["cosmo"])
        self.__dict__.update(d)

    @time_per_day.validator
    def _time_per_day_vld(self, att, val):
        day_length = 24 * un.hour if self.observatory.world == "earth" else 655.2 * un.hour

        if not 0 * un.hour <= val <= day_length:
            raise ValueError(f"time_per_day should be between 0 and {day_length}")

    @track.validator
    def _track_vld(self, att, val):
        if val is not None:
            day_length = 24 * un.hour if self.observatory.world == "earth" else 655.2 * un.hour

            if not 0 * un.hour <= val <= day_length:
                raise ValueError(f"track should be between 0 and {day_length}")

    @lst_bin_size.validator
    def _lst_bin_size_vld(self, att, val):
        day_length = 24 * un.hour if self.observatory.world == "earth" else 655.2 * un.hour

        if not 0 * un.hour <= val <= day_length:
            raise ValueError(f"lst_bin_size should be between 0 and {day_length}")

        if val > self.time_per_day:
            raise ValueError("lst_bin_size must be <= time_per_day")

    @integration_time.validator
    def _integration_time_vld(self, att, val):
        if val > self.lst_bin_size:
            raise ValueError("integration_time must be <= lst_bin_size")

    @time_per_day.default
    def _time_per_day_default(self):
        if self.observatory.world == "earth":
            return 6 * un.hour
        else:
            return 163.8 * un.hour

    @lst_bin_size.default
    def _lst_bin_size_default(self):
        # time it takes the sky to drift through beam FWHM
        if self.track is not None:
            return self.track
        else:
            return self.observatory.observation_duration

    @n_days.default
    def _n_days_default(self):
        if self.observatory.world == "earth":
            return 180
        else:
            return 6

    @phase_center_dec.default
    def _phase_center_dec_default(self):
        return self.observatory.latitude

    @cached_property
    def beam_crossing_time(self):
        """The time it takes for a source to cross the beam.

        This defines the number of independent cosmic fields observed in a night, which
        in turn defines the cosmic sample variance. Observations are not assumed to be
        averaged coherently within this time (for thermal variance calculations).
        """
        return self.observatory.observation_duration

    @cached_property
    def baseline_groups(
        self,
    ) -> dict[tuple[float, float, float], list[tuple[int, int]]]:
        """A dictionary of redundant baseline groups.

        Keys are tuples of floats (X,Y,LENGTH), and
        values are lists of two-tuples of baseline antenna indices in that particular
        baseline group.
        """
        return self.observatory.get_redundant_baselines(
            baseline_filters=self.baseline_filters, ndecimals=self.redundancy_tol
        )

    def __getstate__(self):
        """Get state so that defaultdict is not used."""
        d = dict(self.__dict__.items())
        if "baseline_groups" in d:
            d["baseline_groups"] = dict(d["baseline_groups"])
        return d

    def __setstate__(self, state):
        """Set state so that defaultdict is restored."""
        if "baseline_groups" in state:
            state["baseline_groups"] = defaultdict(list, state["baseline_groups"])
        self.__dict__.update(state)

    @cached_property
    def baseline_group_coords(self) -> un.Quantity[un.m]:
        """Co-ordinates of baseline groups in metres."""
        return self.observatory.baseline_coords_from_groups(self.baseline_groups)

    @cached_property
    def baseline_group_counts(self) -> np.ndarray:
        """The number of baselines in each group."""
        return self.observatory.baseline_weights_from_groups(self.baseline_groups)

    @cached_property
    def baseline_group_lengths(self) -> un.Quantity[un.m]:
        """The displacement magnitude of the baseline groups."""
        return np.sqrt(np.sum(self.baseline_group_coords**2, axis=1))

    @cached_property
    def bl_min(self) -> un.Quantity[un.m]:
        """Shortest included baseline."""
        return self.baseline_group_lengths.min()

    @cached_property
    def bl_max(self) -> un.Quantity[un.m]:
        """Shortest included baseline."""
        return self.baseline_group_lengths.max()

    @property
    def frequency(self) -> un.Quantity[un.MHz]:
        """Frequency of the observation."""
        return self.observatory.frequency

    @cached_property
    def uv_coverage(self) -> np.ndarray:
        # sourcery skip: assign-if-exp, swap-if-expression
        """A 2D array specifying the effective number of baselines in a grid of UV.

        Defined after earth rotation synthesis for a particular LST bin.
        The u-values on each side of the grid are given by :func:`ugrid`.
        """
        return self.observatory.grid_baselines(
            coherent=self.coherent,
            baselines=self.baseline_group_coords,
            weights=self.baseline_group_counts,
            integration_time=self.integration_time,
            observation_duration=self.lst_bin_size,
            ndecimals=self.redundancy_tol,
            phase_center_dec=self.phase_center_dec,
        )

    @cached_property
    def n_lst_bins(self) -> float:
        """
        Number of LST bins in the complete observation.

        An LST bin is considered a chunk of time that may be averaged
        coherently, so this is given by `hours_per_day/lst_bin_size`,
        where `lst_bin_size` is the time it takes for a source to travel
        through the beam FWHM.
        """
        return (self.time_per_day / self.lst_bin_size).to("").value

    @cached_property
    def Tsky(self) -> un.Quantity[un.K]:
        """Temperature of the sky at the default frequency."""
        return self.tsky_amplitude.to("K") * (self.frequency / self.tsky_ref_freq) ** (
            -self.spectral_index
        )

    @cached_property
    def Trcv(self) -> un.Quantity[un.K]:
        """Receiver temperature."""
        if isinstance(self.observatory.Trcv, un.Quantity):
            return self.observatory.Trcv.to("K")
        else:
            return self.observatory.Trcv(self.frequency).to("K")

    @cached_property
    def Tsys(self) -> un.Quantity[un.K]:
        """System temperature (i.e. Tsky + Trcv)."""
        return self.Tsky.to("K") + self.Trcv

    @cached_property
    def redshift(self) -> float:
        """Central redshift of the observation."""
        return conv.f2z(self.frequency)

    @cached_property
    def eta(self) -> un.Quantity[1 / un.MHz]:
        """The fourier dual of the frequencies of the observation."""
        return np.fft.fftfreq(self.n_channels, self.bandwidth.to("MHz") / self.n_channels)

    @cached_property
    def kparallel(self) -> un.Quantity[un.littleh / un.Mpc]:
        """1D array of kpar values, defined by the bandwidth and number of channels.

        Order of the values is the same as `fftfreq` (i.e. zero-first)
        """
        return (
            conv.dk_deta(self.redshift, self.cosmo, approximate=self.use_approximate_cosmo)
            * self.eta
        )

    @cached_property
    def total_integration_time(self) -> un.Quantity[un.s]:
        """The total (effective) integration time over UV bins for a particular LST bin.

        The u-values on each side of the grid are given by :func:`ugrid`.
        """
        return self.uv_coverage * self.n_days * self.integration_time.to(un.s)

    @cached_property
    def Trms(self) -> un.Quantity[un.K]:
        """Effective radiometric noise temperature per UV bin.

        (i.e. divided by bandwidth and integration time).
        The u-values on each side of the grid are given by :func:`ugrid`.
        """
        out = np.ones(self.total_integration_time.shape) * np.inf * self.Tsys.unit
        mask = self.total_integration_time > 0
        out[mask] = self.Tsys.to("K") / np.sqrt(
            2 * self.bandwidth * self.total_integration_time[mask]
        ).to("")
        return out

    @cached_property
    def ugrid(self) -> np.ndarray:
        """Centres of the linear grid which defines a side of the UV grid.

        The UV grid is defined by :func:`uv_coverage`.
        """
        return self.observatory.ugrid(self.bl_max)

    @cached_property
    def ugrid_edges(self) -> np.ndarray:
        """Edges of the linear grid which defines a side of the UV grid.

        The UV grid is defined by :func:`uv_coverage`.
        """
        return self.observatory.ugrid_edges(self.bl_max)

    def clone(self, **kwargs) -> Observation:
        """Create a clone of this instance, with arbitrary changes to parameters."""
        return attr.evolve(self, **kwargs)



"""
Classes from which sensitivities can be obtained.

This module modularizes the previous version's `calc_sense.py`, and enables
multiple sensitivity kinds to be defined. By default, a :class:`PowerSpectrum`
sensitivity class is provided, which offers the same results as previous versions.
In the future, we may provide things like ``ImagingSensitivity`` or
``WaveletSensitivity`` for example.
"""

from __future__ import annotations

import importlib
import logging
from collections.abc import Mapping
from functools import cached_property
from os import path
from pathlib import Path
from typing import Callable

import attr
import h5py
import hickle
import numpy as np
import tqdm
from astropy import units as un
from astropy.cosmology import LambdaCDM
from astropy.cosmology.units import littleh, with_H0
from astropy.io.misc import yaml
from attr import validators as vld
from hickleable import hickleable
from methodtools import lru_cache

from . import _utils as ut
from . import config
from . import conversions as conv
from . import observation as obs
from . import units as tp
from .theory import _ALL_THEORY_POWER_SPECTRA, EOS2021, TheoryModel

logger = logging.getLogger(__name__)


@hickleable(evaluate_cached_properties=True)
@attr.s(kw_only=True)
class Sensitivity:
    """
    Base class for sensitivity calculations.

    Parameters
    ----------
    observation : :class:`~py21cmsense.Observation` instance
        An object defining the observatory and observation used to derive sensitivity.
    no_ns_baselines : bool, optional
        Remove pure north/south baselines (u=0) from the sensitivity calculation.
        These baselines can potentially have higher systematics, so excluding them
        represents a conservative choice.
    """

    observation: obs.Observation = attr.ib(validator=vld.instance_of(obs.Observation))
    no_ns_baselines: bool = attr.ib(default=False, converter=bool)

    @staticmethod
    def _load_yaml(yaml_file):
        if isinstance(yaml_file, str):
            with open(yaml_file) as fl:
                data = yaml.load(fl)
        elif isinstance(yaml_file, Mapping):
            data = yaml_file
        else:
            raise ValueError("yaml_file must be a string filepath or a raw dict from such a file.")
        return data

    @classmethod
    def from_yaml(cls, yaml_file) -> Sensitivity:
        """Construct a :class:`Sensitivity` object from a YAML configuration."""
        data = cls._load_yaml(yaml_file)

        klass = data.pop("class", cls)

        if isinstance(yaml_file, str) and not path.isabs(data["observation"]):
            obsfile = path.join(path.dirname(yaml_file), data.pop("observation"))
        else:
            obsfile = data.pop("observation")

        if obsfile.endswith(".yml"):
            observation = obs.Observation.from_yaml(obsfile)
        elif h5py.is_hdf5(obsfile):
            observation = hickle.load(obsfile)
        else:
            raise ValueError("observation must be a filename with extension .yml or .h5")

        return klass(observation=observation, **data)

    def clone(self, **kwargs):
        """Clone the object with new parameters."""
        return attr.evolve(self, **kwargs)

    def at_frequency(self, frequency: un.Quantity[un.MHz]) -> Sensitivity:
        """Return a new object at a new frequency."""
        return self.clone(
            observation=self.observation.clone(
                observatory=self.observation.observatory.clone(
                    beam=self.observation.observatory.beam.clone(frequency=frequency)
                )
            )
        )

    @property
    def cosmo(self) -> LambdaCDM:
        """The cosmology to use in the sensitivity calculations."""
        return self.observation.cosmo

    @property
    def frequency(self) -> un.Quantity[un.MHz]:
        """The frequency of the observation."""
        return self.observation.frequency


@attr.s(kw_only=True)
class PowerSpectrum(Sensitivity):
    """
    A Power Spectrum sensitivity calculator.

    Note that the sensitivity calculation contains two major parts: thermal variance
    and sample variance (aka cosmic variance). The latter requires a model of the power
    spetrum itself, which you should provide via ``k_21`` and ``delta_21``.
    Remember that the power spectrum is redshift dependent, and so should be supplied
    differently at each frequency being calculated.

    Parameters
    ----------
    horizon_buffer : float or Quantity
        A buffer to add to the horizon line in order to excise foreground-contaminated modes.
    foreground_model : str, {moderate, optimistic}
        Which approach to take for foreground excision. Moderate uses a defined horizon buffer,
        while optimistic excludes all k modes inside the primary field of view.
    k_21 : array or Quantity, optional
        An array of wavenumbers used to define a cosmological power spectrum, in order to get
        sample variance. If not a Quantity, will assume k has units of 1/Mpc, though it will
        convert these units to h/Mpc throughout the class. Default is to use built-in
        data file from 21cmFAST.
    delta_21 : array or Quantity, optional
        An array of Delta^2 power spectrum values used for sample variance.
        If not a Quantity, will assume units of mK^2.
    systematics_mask : callable
        A function that takes a single kperp and an array of kpar, and returns a boolean
        array specifying which of the k's are useable after accounting for systematics.
        that is, it returns False for k's affected by systematics.
    """

    horizon_buffer: tp.Wavenumber = attr.ib(default=0.1 * littleh / un.Mpc)
    foreground_model: str = attr.ib(
        default="moderate", validator=vld.in_(["moderate", "optimistic", "foreground_free"])
    )
    theory_model: TheoryModel = attr.ib()

    systematics_mask: Callable | None = attr.ib(None)

    @horizon_buffer.validator
    def _horizon_buffer_validator(self, att, val):
        tp.vld_unit(littleh / un.Mpc, with_H0(self.cosmo.H0))(self, att, val)
        ut.nonnegative(self, att, val)

    @classmethod
    def from_yaml(cls, yaml_file) -> Sensitivity:
        """
        Construct a PowerSpectrum sensitivity from yaml.

        YAML spec has p21 as a file which obtains k_21 and delta_21.
        It is assumed that k in the file is in units of 1/Mpc (true for 21cmFAST).
        """
        data = cls._load_yaml(yaml_file)

        if "plugins" in data:
            if not isinstance(data["plugins"], list):
                raise ValueError(
                    "plugins in YAML file must be a list of modules."
                )  # pragma: no cover

            for mdl in data.pop("plugins"):
                try:
                    importlib.import_module(mdl)
                except Exception as e:  # noqa: PERF203
                    raise ImportError(f"Could not import {mdl}") from e

        if "theory_model" in data:
            data["theory_model"] = _ALL_THEORY_POWER_SPECTRA[data["theory_model"]]()

        if isinstance(yaml_file, str):
            obsfile = path.join(path.dirname(yaml_file), data.pop("observation"))
        else:
            obsfile = data.pop("observation")

        data["observation"] = obsfile

        return super().from_yaml(data)

    @theory_model.default
    def _theory_model_default(self):
        return EOS2021()

    @theory_model.validator
    def _theory_model_validator(self, att, val):
        if not isinstance(val, TheoryModel):
            raise ValueError("The theory_model must be an instance of TheoryModel")

    @cached_property
    def k1d(self) -> tp.Wavenumber:
        """1D array of wavenumbers for which sensitivities will be generated."""
        delta = (
            conv.dk_deta(
                self.observation.redshift,
                self.cosmo,
                approximate=self.observation.use_approximate_cosmo,
            )
            / self.observation.bandwidth
        )
        dv = delta.value
        return np.arange(dv, dv * self.observation.n_channels, dv) * delta.unit

    @cached_property
    def X2Y(self) -> un.Quantity[un.Mpc**3 / littleh**3 / un.steradian / un.GHz]:
        """Cosmological scaling factor X^2*Y (eg. Parsons 2012)."""
        return conv.X2Y(
            self.observation.redshift,
            approximate=self.observation.use_approximate_cosmo,
        )

    @cached_property
    def uv_coverage(self) -> np.ndarray:
        """The UV-coverage of the array, with unused/redundant baselines set to zero."""
        grid = self.observation.uv_coverage.copy()
        size = grid.shape[0]

        # Cut unnecessary data out of uv coverage: auto-correlations & half of uv
        # plane (which is not statistically independent for real sky)
        grid[size // 2, size // 2] = 0.0
        grid[:, : size // 2] = 0.0
        grid[size // 2 :, size // 2] = 0.0

        if self.no_ns_baselines:
            grid[:, size // 2] = 0.0

        return grid

    def power_normalisation(self, k: tp.Wavenumber) -> float:
        """Normalisation constant for power spectrum."""
        assert hasattr(k, "unit")
        assert k.unit.is_equivalent(littleh / un.Mpc)

        return (
            self.X2Y
            * self.observation.observatory.beam.b_eff
            * self.observation.bandwidth
            * k**3
            / (2 * np.pi**2)
        ).to_value("")

    def thermal_noise(
        self, k_par: tp.Wavenumber, k_perp: tp.Wavenumber, trms: tp.Temperature
    ) -> tp.Delta:
        """Thermal noise contribution at particular k mode."""
        k = np.sqrt(k_par**2 + k_perp**2)
        scalar = self.power_normalisation(k)
        return scalar * trms.to("mK") ** 2

    def sample_noise(self, k_par: tp.Wavenumber, k_perp: tp.Wavenumber) -> tp.Delta:
        """Sample variance contribution at a particular k mode."""
        k = np.sqrt(k_par**2 + k_perp**2).to_value(
            littleh / un.Mpc if self.theory_model.use_littleh else un.Mpc**-1,
            with_H0(self.cosmo.H0),
        )
        return self.theory_model.delta_squared(self.observation.redshift, k)

    @cached_property
    def _nsamples_2d(
        self,
    ) -> dict[str, dict[tp.Wavenumber, un.Quantity[1 / un.mK**4]]]:
        """Mid-way product specifying thermal and sample variance over the 2D grid."""
        # set up blank arrays/dictionaries
        sense = {"sample": {}, "thermal": {}, "both": {}}

        # loop over uv_coverage to calculate k_pr
        nonzero = np.where(self.uv_coverage > 0)
        for iu, iv in tqdm.tqdm(
            zip(nonzero[1], nonzero[0]),
            desc="calculating 2D sensitivity",
            unit="uv-bins",
            disable=not config.PROGRESS,
            total=len(nonzero[1]),
        ):
            u, v = self.observation.ugrid[iu], self.observation.ugrid[iv]
            trms = self.observation.Trms[iv, iu]

            if np.isinf(trms):
                continue

            umag = np.sqrt(u**2 + v**2)
            k_perp = umag * conv.dk_du(
                self.observation.redshift,
                self.cosmo,
                approximate=self.observation.use_approximate_cosmo,
            )

            hor = self.horizon_limit(umag)

            if k_perp not in sense["thermal"]:
                sense["thermal"][k_perp] = np.zeros(len(self.observation.kparallel)) / un.mK**4
                sense["sample"][k_perp] = np.zeros(len(self.observation.kparallel)) / un.mK**4
                sense["both"][k_perp] = np.zeros(len(self.observation.kparallel)) / un.mK**4

            # Exclude parallel modes dominated by foregrounds
            kpars = self.observation.kparallel[self.observation.kparallel >= hor]

            if not len(kpars):
                continue

            start = np.where(self.observation.kparallel >= hor)[0][0]
            n_inds = (self.observation.kparallel.size - 1) // 2 + 1
            inds = np.arange(start=start, stop=n_inds)

            thermal = self.thermal_noise(kpars, k_perp, trms)
            sample = self.sample_noise(kpars, k_perp)

            # The following assumes that the power spectra are averaged with inverse
            # variance weighting.
            t = 1.0 / thermal**2
            s = 1.0 / sample**2
            ts = 1.0 / (thermal + sample) ** 2
            sense["thermal"][k_perp][inds] += t
            sense["thermal"][k_perp][-inds] += t
            sense["sample"][k_perp][inds] += s
            sense["sample"][k_perp][-inds] += s
            sense["both"][k_perp][inds] += ts
            sense["both"][k_perp][-inds] += ts

        return sense

    @lru_cache()
    def calculate_sensitivity_2d(
        self, thermal: bool = True, sample: bool = True
    ) -> dict[tp.Wavenumber, tp.Delta]:
        """
        Calculate power spectrum sensitivity for a grid of cylindrical k modes.

        Parameters
        ----------
        thermal : bool, optional
            Whether to calculate thermal contribution to the sensitivity
        sample : bool, optional
            Whether to calculate sample variance contribution to sensitivity.

        Returns
        -------
        dict :
            Keys are cylindrical kperp values and values are arrays aligned with
            `observation.kparallel`, defining uncertainty in mK^2.
        """
        if thermal and sample:
            logger.info("Getting Combined Variance")
            sense = self._nsamples_2d["both"]
        elif thermal:
            logger.info("Getting Thermal Variance")
            sense = self._nsamples_2d["thermal"]
        elif sample:
            logger.info("Getting Sample Variance")
            sense = self._nsamples_2d["sample"]
        else:
            raise ValueError("Either thermal or sample must be True")

        # errors were added in inverse quadrature, now need to invert and take
        # square root to have error bars; also divide errors by number of indep. fields
        final_sense = {}
        for k_perp in sense:
            mask = sense[k_perp] > 0
            if self.systematics_mask is not None:
                mask &= self.systematics_mask(k_perp, self.observation.kparallel)

            if not np.any(mask):
                continue

            final_sense[k_perp] = np.inf * np.ones(len(mask)) * un.mK**2
            if thermal:
                total_std = thermal_std = 1 / np.sqrt(
                    self._nsamples_2d["thermal"][k_perp][mask] * self.observation.n_lst_bins
                )
            if sample:
                total_std = sample_std = 1 / np.sqrt(
                    self._nsamples_2d["sample"][k_perp][mask]
                    * (self.observation.time_per_day / self.observation.lst_bin_size).to("")
                )
            if thermal and sample:
                total_std = thermal_std + sample_std

            final_sense[k_perp][mask] = total_std

        return final_sense

    def calculate_sensitivity_2d_grid(
        self,
        kperp_edges: tp.Wavenumber,
        kpar_edges: tp.Wavenumber,
        thermal: bool = True,
        sample: bool = True,
    ) -> tp.Delta:
        """Calculate the 2D cylindrical sensitivity on a grid of kperp/kpar.

        Parameters
        ----------
        kperp_edges
            The edges of the bins in kperp.
        kpar_edges
            The edges of the bins in kpar.
        """
        sense2d_inv = np.zeros((len(kperp_edges) - 1, len(kpar_edges) - 1)) << (1 / un.mK**4)
        sense = self.calculate_sensitivity_2d(thermal=thermal, sample=sample)

        assert np.all(np.diff(kperp_edges) > 0)
        assert np.all(np.diff(kpar_edges) > 0)

        for k_perp in tqdm.tqdm(
            sense.keys(),
            desc="averaging to 2D grid",
            unit="kperp-bins",
            disable=not config.PROGRESS,
        ):
            if k_perp < kperp_edges[0] or k_perp >= kperp_edges[-1]:
                continue

            # Get the kperp bin it's in.
            kperp_indx = np.where(k_perp >= kperp_edges)[0][-1]

            kpar_indx = np.digitize(self.observation.kparallel, kpar_edges) - 1
            good_ks = kpar_indx >= 0
            good_ks &= kpar_indx < len(kpar_edges) - 1

            sense2d_inv[kperp_indx][kpar_indx[good_ks]] += 1.0 / sense[k_perp][good_ks] ** 2

        # invert errors and take square root again for final answer
        sense2d = np.ones(sense2d_inv.shape) * un.mK**2 * np.inf
        mask = sense2d_inv > 0
        sense2d[mask] = 1 / np.sqrt(sense2d_inv[mask])
        return sense2d

    def horizon_limit(self, umag: float) -> tp.Wavenumber:
        """
        Calculate a horizon limit, with included buffer, if appropriate.

        Parameters
        ----------
        umag : float
            Baseline length (in wavelengths) at which to compute the horizon limit.

        Returns
        -------
        float :
            Horizon limit, in h/Mpc.
        """
        horizon = (
            conv.dk_deta(
                self.observation.redshift,
                self.cosmo,
                approximate=self.observation.use_approximate_cosmo,
            )
            * umag
            / self.observation.frequency
        )
        # calculate horizon limit for baseline of length umag
        if self.foreground_model in ["moderate", "pessimistic"]:
            return horizon + self.horizon_buffer
        elif self.foreground_model in ["optimistic"]:
            return horizon * np.sin(self.observation.observatory.beam.first_null / 2)
        elif self.foreground_model in ["foreground_free"]:
            return 0

    def _average_sense_to_1d(
        self, sense: dict[tp.Wavenumber, tp.Delta], k1d: tp.Wavenumber | None = None
    ) -> tp.Delta:
        """Bin 2D sensitivity down to 1D."""
        if k1d is None:
            k1d = self.k1d
        sense1d_inv = np.zeros(len(k1d)) / un.mK**4

        for k_perp in tqdm.tqdm(
            sense.keys(),
            desc="averaging to 1D",
            unit="kperp-bins",
            disable=not config.PROGRESS,
        ):
            k = np.sqrt(self.observation.kparallel**2 + k_perp**2)

            good_ks = k >= k1d.min()
            good_ks &= k < k1d.max()

            for cnt, kbin in enumerate(ut.find_nearest(k1d, k[good_ks])):
                sense1d_inv[kbin] += 1.0 / sense[k_perp][good_ks][cnt] ** 2

        # invert errors and take square root again for final answer
        sense1d = np.ones(sense1d_inv.shape) * un.mK**2 * np.inf
        mask = sense1d_inv > 0
        sense1d[mask] = 1 / np.sqrt(sense1d_inv[mask])
        return sense1d

    @lru_cache()
    def calculate_sensitivity_1d(self, thermal: bool = True, sample: bool = True) -> tp.Delta:
        """Calculate a 1D sensitivity curve.

        Parameters
        ----------
        thermal
            Whether to calculate thermal contribution to the sensitivity
        sample
            Whether to calculate sample variance contribution to sensitivity.

        Returns
        -------
        array :
            1D array with units mK^2... the variance of spherical k modes.
        """
        sense = self.calculate_sensitivity_2d(thermal=thermal, sample=sample)
        return self._average_sense_to_1d(sense)

    def calculate_sensitivity_1d_binned(self, k: tp.Wavenumber, **kwargs):
        """Calculate the 1D sensitivity at arbitrary k-bins."""
        sense2d = self.calculate_sensitivity_2d(**kwargs)
        return self._average_sense_to_1d(sense2d, k1d=k)

    @property
    def delta_squared(self) -> tp.Delta:
        """The fiducial 21cm power spectrum evaluated at :attr:`k1d`."""
        k = self.k1d.to_value(
            littleh / un.Mpc if self.theory_model.use_littleh else un.Mpc**-1,
            with_H0(self.cosmo.H0),
        )
        return self.theory_model.delta_squared(self.observation.redshift, k)

    @lru_cache()
    def calculate_significance(self, thermal: bool = True, sample: bool = True) -> float:
        """
        Calculate significance of a detection of the default cosmological power spectrum.

        Parameters
        ----------
        thermal : bool, optional
            Whether to calculate thermal contribution to the sensitivity
        sample : bool, optional
            Whether to calculate sample variance contribution to sensitivity.

        Returns
        -------
        float :
            Significance of detection (in units of sigma).
        """
        sense1d = self.calculate_sensitivity_1d(thermal=thermal, sample=sample)

        snr = self.delta_squared / sense1d
        return np.sqrt(float(np.dot(snr, snr.T)))

    def plot_sense_2d(self, sense2d: dict[tp.Wavenumber, tp.Delta]):
        # sourcery skip: raise-from-previous-error
        """Create a colormap plot of the sensitivity un UV bins."""
        try:
            import matplotlib.pyplot as plt
        except ImportError as e:  # pragma: no cover
            raise ImportError("matplotlib is required to make plots...") from e

        keys = sorted(sense2d.keys())
        x = np.array([v.value for v in keys])
        x = (
            np.repeat(x, len(self.observation.kparallel))
            .reshape((len(x), len(self.observation.kparallel)))
            .T
        )
        y = np.fft.fftshift(
            np.repeat(self.observation.kparallel.value, x.shape[1]).reshape(
                (len(self.observation.kparallel), x.shape[1])
            )
        )
        z = np.array([np.fft.fftshift(sense2d[key]) for key in keys]).T

        plt.pcolormesh(x, y, np.log10(z))
        cbar = plt.colorbar()
        cbar.set_label(r"$\log_{10} \delta \Delta^2\ \  [{\rm mK}^2]$", fontsize=14)
        plt.xlabel(r"$k_\perp$ [h/Mpc]", fontsize=14)
        plt.ylabel(r"$k_{||}$ [h/Mpc]", fontsize=14)

    def write(
        self,
        filename: str | Path,
        thermal: bool = True,
        sample: bool = True,
        prefix: str | None = None,
        direc: str | Path = ".",
    ) -> Path:
        """Save sensitivity results to HDF5 file.

        Returns
        -------
        filename
            The path to the file that is written.
        """
        out = self._get_all_sensitivity_combos(thermal, sample)
        prefix = f"{prefix}_" if prefix else ""
        if filename is None:
            filename = Path(
                f"{prefix}{self.foreground_model}_{self.observation.frequency:.3f}.h5".replace(
                    " ", ""
                )
            )
        else:
            filename = Path(filename)

        if direc is not None:
            filename = Path(direc) / filename

        logger.info(f"Writing sensitivies to '{filename}'")
        with h5py.File(filename, "w") as fl:
            # TODO: We should be careful to try and write everything into this file
            # i.e. all the parameters etc.

            for k, v in out.items():
                fl[k] = v
                fl[k.replace("noise", "snr")] = self.delta_squared / v

            fl["k"] = self.k1d.to("1/Mpc", with_H0(self.cosmo.H0)).value
            fl["delta_squared"] = self.delta_squared

            fl.attrs["total_snr"] = self.calculate_significance()
            fl.attrs["foreground_model"] = self.foreground_model
            fl.attrs["horizon_buffer"] = self.horizon_buffer
            fl.attrs["k_unit"] = "1/Mpc"

        return filename

    def plot_sense_1d(self, sample: bool = True, thermal: bool = True):
        """Create a plot of the sensitivity in 1D k-bins."""
        try:
            import matplotlib.pyplot as plt
        except ImportError as e:  # pragma: no cover
            raise ImportError("matplotlib is required to make plots...") from e

        out = self._get_all_sensitivity_combos(thermal, sample)
        for key, value in out.items():
            plt.plot(self.k1d, value, label=key)
            plt.xscale("log")
            plt.yscale("log")
            plt.xlabel("k [h/Mpc]")
            plt.ylabel(r"$\Delta^2_N \  [{\rm mK}^2]$")
            plt.legend()
            plt.title(f"z={conv.f2z(self.observation.frequency):.2f}")

        return plt.gcf()

    def _get_all_sensitivity_combos(self, thermal: bool, sample: bool) -> dict[str, tp.Delta]:
        result = {}
        if thermal:
            result["thermal_noise"] = self.calculate_sensitivity_1d(sample=False)
        if sample:
            result["sample_noise"] = self.calculate_sensitivity_1d(thermal=False, sample=True)

        if thermal and sample:
            result["sample+thermal_noise"] = self.calculate_sensitivity_1d(
                thermal=True, sample=True
            )

        return result



"""A package for calculate sensitivies of 21-cm interferometers."""

from importlib.metadata import PackageNotFoundError, version

try:
    __version__ = version("21cmSense")
except PackageNotFoundError:  # pragma: no cover
    __version__ = "unknown"
finally:
    del version, PackageNotFoundError

__all__ = [
    "BaselineRange",
    "GaussianBeam",
    "Observation",
    "Observatory",
    "PowerSpectrum",
    "data",
    "hera",
    "theory",
    "yaml",
]
from . import data, theory, yaml
from .antpos import hera
from .baseline_filters import BaselineRange
from .beam import GaussianBeam
from .observation import Observation
from .observatory import Observatory
from .sensitivity import PowerSpectrum



"""
Common 21 cm conversions.

Provides conversions between observing co-ordinates and cosmological co-ordinates.
"""

from __future__ import annotations

import numpy as np
from astropy import constants as cnst
from astropy import units as un
from astropy.cosmology import FLRW, Planck15
from astropy.cosmology.units import littleh

from . import units as tp

# The frequency of the 21cm line emission.
f21 = 1.42040575177 * un.GHz


def f2z(fq: tp.Frequency) -> float:
    """
    Convert frequency to redshift for 21 cm line.

    Parameters
    ----------
    fq : float or astropy.Quantity
        If float, it is interpreted as being in GHz.

    Returns
    -------
    dimensionless astropy.Quantity : The redshift
    """
    return float(f21 / fq - 1)


def z2f(z: float | np.array) -> un.Quantity[un.GHz]:
    """
    Convert redshift to z=0 frequency for 21 cm line.

    Parameters
    ----------
    z
        Redshift

    Returns
    -------
    astropy.Quantity : the frequency
    """
    return f21 / (1 + z)


def dL_dth(
    z: float | np.array,
    cosmo: FLRW = Planck15,
    approximate=False,
) -> un.Quantity[un.Mpc / un.rad / littleh]:
    """
    Return the factor to convert radians to transverse distance at redshift z.

    Parameters
    ----------
    z : float
        The redshift

    Returns
    -------
    astropy.Quantity : the factor (in Mpc/h/radian) which converts from an angle
        to a transverse distance.

    Notes
    -----
    From Furlanetto et al. (2006)
    """
    if approximate:
        return (1.9 * (1.0 / un.arcmin) * ((1 + z) / 10.0) ** 0.2).to(1 / un.rad) * un.Mpc / littleh
    else:
        return cosmo.h * cosmo.comoving_transverse_distance(z) / un.rad / littleh


def dL_df(
    z: float | np.array,
    cosmo: FLRW = Planck15,
    approximate=False,
) -> un.Quantity[un.Mpc / un.MHz / littleh]:
    """
    Get the factor to convert bandwidth to line-of-sight distance in Mpc/h.

    Parameters
    ----------
    z : float
        The redshift
    """
    if approximate:
        return (
            (1.7 / 0.1)
            * ((1 + z) / 10.0) ** 0.5
            * (cosmo.Om0 / 0.15) ** -0.5
            * un.Mpc
            / littleh
            / un.MHz
        )
    else:
        return (cosmo.h * cnst.c * (1 + z) / (z2f(z) * cosmo.H(z) * littleh)).to(
            "Mpc/(MHz*littleh)"
        )


def dk_du(
    z: float | np.array,
    cosmo: FLRW = Planck15,
    approximate=False,
) -> un.Quantity[littleh / un.Mpc]:
    """
    Get factor converting bl length in wavelengths to h/Mpc.

    Parameters
    ----------
    z : float
        redshift

    Notes
    -----
    Valid for u >> 1
    """
    # from du = 1/dth, which derives from du = d(sin(th)) using the small-angle approx
    return 2 * np.pi / dL_dth(z, cosmo, approximate=approximate) / un.rad


def dk_deta(
    z: float | np.array,
    cosmo: FLRW = Planck15,
    approximate=False,
) -> un.Quantity[un.MHz * littleh / un.Mpc]:
    """
    Get gactor converting inverse frequency to inverse distance.

    Parameters
    ----------
    z: float
        Redshift
    """
    return 2 * np.pi / dL_df(z, cosmo, approximate=approximate)


def X2Y(
    z: float | np.array,
    cosmo: FLRW = Planck15,
    approximate=False,
) -> un.Quantity[un.Mpc**3 / littleh**3 / un.steradian / un.MHz]:
    """
    Obtain the conversion factor between observing co-ordinates and cosmological volume.

    Parameters
    ----------
    z: float
        Redshift
    cosmo: astropy.cosmology.FLRW instance
        A cosmology.

    Returns
    -------
    astropy.Quantity: the conversion factor. Units are Mpc^3/h^3 / (sr MHz).
    """
    return dL_dth(z, cosmo, approximate=approximate) ** 2 * dL_df(z, cosmo, approximate=approximate)



"""Module dealing with types and units throughout the package."""

from __future__ import annotations

from typing import Any, Callable

import attr
from astropy import constants as cnst
from astropy import units as un
from astropy.cosmology.units import littleh, redshift

un.add_enabled_units([littleh, redshift])


class UnitError(ValueError):
    """An error pertaining to having incorrect units."""


Length = un.Quantity["length"]
Meters = un.Quantity["m"]
Time = un.Quantity["time"]
Frequency = un.Quantity["frequency"]
Temperature = un.Quantity["temperature"]
TempSquared = un.Quantity[un.get_physical_type("temperature") ** 2]
Wavenumber = un.Quantity[littleh / un.Mpc]
Delta = un.Quantity[un.mK**2]
Angle = un.Quantity["angle"]

time_as_distance = [
    (
        un.s,
        un.m,
        lambda x: cnst.c.to_value("m/s") * x,
        lambda x: x / cnst.c.to_value("m/s"),
    )
]


def vld_physical_type(unit: str) -> Callable[[Any, attr.Attribute, Any], None]:
    """Attr validator to check physical type."""

    def _check_type(self: Any, att: attr.Attribute, val: Any):
        if not isinstance(val, un.Quantity):
            raise UnitError(f"{att.name} must be an astropy Quantity!")
        if val.unit.physical_type != unit:
            raise un.UnitConversionError(
                f"{att.name} must have physical type of '{unit}'. Got '{val.unit.physical_type}'"
            )

    return _check_type


def vld_unit(unit, equivalencies=()):
    """Attr validator to check unit equivalence."""

    def _check_unit(self, att, val):
        if not isinstance(val, un.Quantity):
            raise UnitError(f"{att.name} must be an astropy Quantity!")

        if not val.unit.is_equivalent(unit, equivalencies):
            raise un.UnitConversionError(f"{att.name} not convertible to {unit}. Got {val.unit}")

    return _check_unit


def _tuplify(x, n: int = 1):
    try:
        return tuple(x)
    except TypeError:
        return (x,) * n
